#
# SharedArray can be initialized in Parent or Child process
# Uses anonymous mmap with tagnames
# A SharedArray can be send of multiprocessing queues
# Pickles only the mmap, tagname, sizes, ... but not the buffer.

import numpy as np
import os, time

import mmap, _winapi
import tempfile
        

class SharedArray:
    """
    Using shared memory.
    If pickled to another process, the memory is not copied.
    """
    
    #part of code copied from multiprocessing.heap module
    
    _rand = tempfile._RandomNameSequence()

    def __init__(self, shape, dtype=float):
        #convert dtype string to real dtype
        self.dtype = np.dtype(dtype)
        self.shape = shape
        for i in range(100):
            name = 'pym-%d-%s' % (os.getpid(), next(self._rand))
            buf = mmap.mmap(-1, self.bytesize, name)
            if _winapi.GetLastError() == 0:
                break
            # We have reopened a preexisting mmap.
            buf.close()
        else:
            raise FileExistsError('Cannot find name for new mmap')
        self.name = name   
        self._ndarray = None
        self._bindex = None
        self.base = buf
                        
    @staticmethod
    def from_ndarray(array):
        sa =  SharedArray(array.shape, array.dtype)
        sa.ndarray[:] = array
        return sa        
        
    @property
    def size(self):
         return np.multiply.reduce(self.shape, dtype='int64')
    
    @property    
    def ndim(self):
        return len(self.shape)
         
    @property
    def bytesize(self):
         return self.size * self.dtype.itemsize

    def __getstate__(self):
        return (self.name, self.dtype, self.shape)
        
    def __setstate__(self, state):    
        self.name, self.dtype, self.shape = self._state = state             
        self.base = mmap.mmap(-1, self.bytesize, self.name)
        assert _winapi.GetLastError() == _winapi.ERROR_ALREADY_EXISTS
        self._ndarray = None
        
    def _as_ndarray(self):
        arr = np.frombuffer(self.base, self.dtype, self.size).reshape(self.shape)
        return arr
        
    @property
    def ndarray(self):
        if self._ndarray is None:
            self._ndarray = self._as_ndarray()
        return self._ndarray        
            
    def __getitem__(self, slices):
        return self.ndarray.__getitem__(slices)
        
    def __setitem__(self, slices, other):
        return self.ndarray.__setitem__(slices, other)    
            
    def __str__(self):
        return self.ndarray.__str__()
            
    def __repr__(self):
        s = self.ndarray.__repr__()
        s = s.replace('array', 'SharedArray')
        return s
        
    def __dir__(self):
        d = dir(self.ndarray)
        d.extend(self.__dict__.keys())
        return d
        
    def __getattr__(self, attr):
        return getattr(self.ndarray, attr)
        
            


from qtpy import QtCore, QtGui, QtWidgets

class GridSplitter(QtWidgets.QGridLayout):
    
    layoutChanged = QtCore.Signal()
    
    def __init__(self, parent):
        super().__init__(parent=parent)
        self.setSpacing(0)
        self.setContentsMargins(0,0,0,0)
        self.splitters = []
        
    def addWidget(self, widget, row, col, rowspan=1, colspan=1):        
        super().addWidget(widget, row*2, col*2, rowspan*2-1, colspan*2-1)
        
        if (self.itemAtPosition(row*2-1, col*2) == None) and (row > 0):
            splitter = GridSeperator(self.parent(), QtCore.Qt.Vertical)
            splitter.attachGridSplitter(self, row-1, row)
            self.splitters.append(splitter)
            super().addWidget(splitter, row*2-1, col*2)
            
        if (self.itemAtPosition(row*2, col*2-1) == None) and (col > 0):
            splitter = GridSeperator(self.parent(), QtCore.Qt.Horizontal)
            splitter.attachGridSplitter(self, col-1, col)
            self.splitters.append(splitter)
            super().addWidget(splitter, row*2, col*2-1)
        
    def getRowStretches(self):
        """get the row streches from all rows, skip the row splitters"""
        stretches = []
        for i in range(0, self.rowCount(), 2):
            stretches.append(self.cellRect(i, 0).height())
        return stretches
        
    def setRowStretches(self, stretches):
        for (i, stretch) in zip(range(0, self.rowCount(), 2), stretches):            
            self.setRowStretch(i, stretch)
                
    def resetRowStretches(self):
        for i in range(0, self.rowCount()):
            self.setRowStretch(i, 0)
            
    def getColumnStretches(self):
        """get the columns streches from all columns, skip the column splitters"""
        stretches = []
        for i in range(0, self.columnCount(), 2):
            stretches.append(self.cellRect(0, i).width())
        return stretches
        
    def setColumnStretches(self, stretches):
        for (i, stretch) in zip(range(0, self.columnCount(), 2), stretches):            
            self.setColumnStretch(i, stretch)
                
    def resetColumnStretches(self):
        for i in range(0, self.rowCount()):
            self.setColumnStretch(i, 0)
            
    def setGeometry(self, arg):            
        self.layoutChanged.emit()
        super().setGeometry(arg)
        
        
class GridSeperator(QtWidgets.QWidget):
    
    def __init__(self, parent=None, orientation = QtCore.Qt.Vertical):
        super().__init__(parent=parent)
        
        self.orientation = orientation
        if orientation == QtCore.Qt.Vertical:
            self.setCursor(QtGui.QCursor(QtCore.Qt.SplitVCursor))
            self.setFixedHeight(5)
        else:
            self.setCursor(QtGui.QCursor(QtCore.Qt.SplitHCursor))
            self.setFixedWidth(5)                        
        self.dragStartX = 0
        self.dragStartY = 0
        
    def mousePressEvent(self, event):
        if event.buttons() == QtCore.Qt.LeftButton:
            self.dragStartX = event.pos().x()
            self.dragStartY = event.pos().y()
            self.row_stretches = self.gridSplitter.getRowStretches()
            self.column_stretches = self.gridSplitter.getColumnStretches()
            
    def mouseMoveEvent(self, event):
        self.movedX = event.pos().x() - self.dragStartX
        self.movedY = event.pos().y() - self.dragStartY
        if self.orientation == QtCore.Qt.Vertical:
            if self.movedY != 0:
                self.restretchRows(self.movedY)

        if self.orientation == QtCore.Qt.Horizontal:
            if self.movedX != 0:
                self.restretchColumns(self.movedX)
            
    def restretchRows(self, moved):
        stretches = self.row_stretches
        stretches[self.lower] += moved
        stretches[self.current] -= moved
        self.gridSplitter.setRowStretches(stretches)
            
    def restretchColumns(self, moved):
        stretches = self.column_stretches
        stretches[self.lower] += moved
        stretches[self.current] -= moved
        self.gridSplitter.setColumnStretches(stretches)           
        
    def attachGridSplitter(self, gridSplitter, lower, current):
        self.gridSplitter = gridSplitter
        self.lower = lower
        self.current = current        
  


"""
The About window of this GUI
"""
import pathlib

from qtpy import QtGui, QtWidgets
from qtpy.QtCore import Qt

from .. import config, __version__, __release__, PROGNAME

respath = pathlib.Path(config['respath'])

class TextEditLinks(QtWidgets.QTextBrowser):
    """
    QTextBrowser used to show clickable links
    """
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.setOpenExternalLinks(True)

    def addLink(self, address, text):
        """addLink"""
        self.append(f'<a href="{address}">{text}</a>')

    def addMail(self, mail_address, text):
        """addMail"""
        self.append(f'<a href="mailto:{mail_address}">{text}</a>')


class AboutScreen(QtWidgets.QDialog):
    """
    The About window
    """
    def __init__(self):
        super().__init__()
        self.initUI()

    def initUI(self):
        """
        Init the GUI items on this widget
        """
        self.setMinimumWidth(500)
        self.setMinimumHeight(400)

        self.setWindowTitle('About')
        self.setWindowFlags(Qt.WindowStaysOnTopHint)

        logopixmap = QtGui.QPixmap(str(respath / 'logo' / 'logo_128px.png'))

        logo = QtWidgets.QLabel()
        logo.setPixmap(logopixmap)

        message = f"""<center><h2>{PROGNAME}</h2>
            <p>Version {__release__}</p>
            """

        cont = TextEditLinks(self)
        cont.setReadOnly(True)
        cont.append(message)

        cont.addLink("http://www.apache.org/licenses/LICENSE-2.0",
            "www.apache.org/licenses/LICENSE-2.0")
        cont.addLink("http://www.fatcow.com/free-icons",
            "www.fatcow.com/free-icons")
        cont.addMail("thomas.cools@telenet.be",
            "thomas.cools@telenet.be")

        hboxlogo = QtWidgets.QHBoxLayout()
        hboxlogo.addWidget(logo)
        hboxlogo.addWidget(cont)

        okButton = QtWidgets.QPushButton("OK")
        hboxbut = QtWidgets.QHBoxLayout()
        hboxbut.addStretch(1)
        hboxbut.addWidget(okButton)

        vbox = QtWidgets.QVBoxLayout()
        vbox.addLayout(hboxlogo)
        vbox.addLayout(hboxbut)

        self.setLayout(vbox)

        okButton.clicked.connect(self.close)

        self.show()



"""Function tools."""
from types import CodeType, FunctionType


def find_nested_func(parent, child_name, globs, closure=None):
    """
    Find nested funcion.

    Return the function named <child_name> that is defined inside
    a <parent> function
    Returns None if nonexistent
    """
    consts = parent.__code__.co_consts
    for item in consts:
        if isinstance(item, CodeType) and item.co_name == child_name:
            return FunctionType(item, globs, closure=closure)



import numpy as np

def getOptimalMinimumSpacing(minVal, maxVal, scale=1):
    dif = abs(maxVal - minVal)
    if dif == 0:
        return 0
    size = dif * scale
    # decide optimal minor tick spacing in pixels (this is just aesthetics)
    pixelSpacing = np.log(size+10) * 5
    #pixelSpacing = np.log(size+10) * 2
    optimalTickCount = size / pixelSpacing
    if optimalTickCount < 1:
        optimalTickCount = 1
    # optimal minor tick spacing 
    minimumSpacing = dif / optimalTickCount  
    return minimumSpacing * scale

def tickSpacing(minimumSpacing=1, noDecimals=False):
    """Return values describing the desired spacing and offset of ticks.
    
    This method is called whenever the axis needs to be redrawn and is a 
    good method to override in subclasses that require control over tick locations.
    
    The return value must be a list of three tuples::
    
        [
            (major tick spacing, offset),
            (minor tick spacing, offset),
            (micro tick spacing, offset),
            ...
        ]
    """
    
    # the largest power-of-10 spacing which is smaller than optimal
    p10unit = 10 ** np.floor(np.log10(minimumSpacing))
    
    if noDecimals and p10unit < 1:
        return [5, 1, 1]
        # (5, 0),
        # (1, 0),
        # (1, 0)
        # ]  
        
    microInternals = np.array([0.2, 0.4, 1., 2.]) * p10unit
    minorInternals = np.array([1., 2., 5., 10.]) * p10unit
    majorInternals = np.array([2., 4., 10., 20.]) * p10unit        
        
        
    # Determine major/minor tick spacings which flank the optimal spacing.
    minorIndex = 3
    while minorInternals[minorIndex-1] > minimumSpacing:
        minorIndex -= 1

    # return [
        # (majorInternals[minorIndex], 0),
        # (minorInternals[minorIndex], 0),
        # (microInternals[minorIndex], 0)]
        
    return [majorInternals[minorIndex], minorInternals[minorIndex], microInternals[minorIndex]]    


def tickValues(minVal, maxVal, scale, minimumSpacing=None, noDecimals=False):
    """
    Return the values and spacing of ticks to draw::
    
        [  
            (spacing, [major ticks]), 
            (spacing, [minor ticks]), 
            ... 
        ]
    
    By default, this method calls tickSpacing to determine the correct tick locations.
    This is a good method to override in subclasses.
    """
    minVal, maxVal = sorted((minVal, maxVal))
    
    if minimumSpacing == None:
        minimumSpacing = getOptimalMinimumSpacing(minVal, maxVal, scale)
                
    ticks = []
    tickLevels = tickSpacing(minimumSpacing / scale, noDecimals)
    allValues = np.array([])
    # for i in range(len(tickLevels)):
        # spacing, offset = tickLevels[i]
    offset = 0
    for spacing in tickLevels:        
        # determine starting tick
        start = (np.ceil((minVal-offset) / spacing) * spacing) + offset
        
        # determine number of ticks
        num = int((maxVal-start) / spacing) + 1
        values = np.arange(num) * spacing + start
        # remove any ticks that were present in higher levels
        # we assume here that if the difference between a tick value and a previously seen tick value
        # is less than spacing/100, then they are 'equal' and we can ignore the new tick.
        values = list(filter(lambda x: all(np.abs(allValues-x) > spacing*0.01), values) )
        allValues = np.concatenate([allValues, values])
        scaledSpacing = spacing * scale
        ticks.append((scaledSpacing, values))        
        
    return ticks


class Ticks:

    def __init__(self, minVal, maxVal, scale, minimumSpacing=None, noDecimals=False):
        if minimumSpacing == None:
            self.minimumSpacing = getOptimalMinimumSpacing(minVal, maxVal, scale)
        else:
            self.minimumSpacing = minimumSpacing
        self.noDecimals = noDecimals
        
        self.values = []
        self.pop_values = []  #to be removed from scene
        self.push_values = [] #to be added to scene
        
        self.update(minVal, maxVal, scale)                       
        
    def update(self, minVal, maxVal, scale):
        self.spacings = tickSpacing(self.minimumSpacing / scale, self.noDecimals)
        
        values = []
        self.pop_values = []  #to be removed from scene
        self.push_values = [] #to be added to scene
        
        allValues = np.array([])
        #for i in range(len(self.tickLevels)):
        for spacing in self.spacings:
            #spacing, offset = self.tickLevels[i]
            
            # determine starting tick
            start = np.ceil(minVal / spacing)
            
            # determine number of ticks
            end = np.floor(maxVal / spacing) + 1
            
            val = np.arange(start, end) * spacing
            
            # remove any ticks that were present in higher levels
            # we assume here that if the difference between a tick value and a previously seen tick value
            # is less than spacing/100, then they are 'equal' and we can ignore the new tick.
            val = list(filter(lambda x: all(np.abs(allValues-x) > spacing*0.01), val) )
            allValues = np.concatenate([allValues, val])
            scaledSpacing = spacing * scale
            values.append((scaledSpacing, val))  
            
        if self.values != []:       
            for i in range(len(values)):            
                old_scaled_spacing, old_values = self.values[i]
                new_scaled_spacing, new_values = values[i]
                push_mask = np.in1d(new_values, old_values, invert = True)
                pop_mask = np.in1d(old_values, new_values, invert = True)
                self.push_values.append(np.array(new_values)[push_mask])
                self.pop_values.append(np.array(old_values)[pop_mask])
        else:
            self.push_values.append(values[0][1])
            self.push_values.append(values[1][1])
            self.push_values.append(values[2][1])
            
        # if len(self.push_values) != 0:
            # print('push', self.push_values)
        
        # if len(self.pop_values) != 0:
            # print('pop', self.pop_values)            
        
        self.values = values



import sys
import threading

from ...core.gui_proxy import GuiProxyBase, StaticGuiCall, gui
from ...core.shellmod import Shell
       
class ConsoleGuiProxy(GuiProxyBase):    
    category = 'console'
    opens_with = ['.py']
    
    def __init__(self):
        pass
        
    def attach(self, gui):
        gui.console = self
        gui.clc = self.clc
        return 'console'
        
    @StaticGuiCall 
    def open(filepath):
        panel = gui.qapp.panels.selected('console')
        ConsoleGuiProxy._gui_execute_file(filepath, panel.panid)

    @StaticGuiCall
    def console(pandid=None, consoletype='thread'):
        """
        Select or create an image panel with id image_id or auto id
        Returns the shared array of the panel and the id.
        
        :param str consoletype: 'main', 'thread', 'child', 'child-thread', 'process'
        """        
        if pandid is None:
            panel = gui.qapp.panels.selected('console')
        else:
            panel = gui.qapp.panels.select_or_new('console', pandid, consoletype)                
            
        return panel.panid    

        
    @StaticGuiCall
    def clc():
        """Clear the Console Output"""      
        # For now, jus take the active console
        # Note that this is not always correct
        # Should first search for the console of the current process and thread
        sys.stdout.flush()
        sys.stderr.flush()
        panel = gui.qapp.panels.selected('console')
        panel.stdio.stdOutputPanel.clear()
        

    @StaticGuiCall
    def text():
        panel = gui.qapp.panels.selected('console')
        text = panel.stdio.stdOutputPanel.toPlainText()
        return text


    def set_mode(self, mode='input', panid=None):
        if panid is None:
            shell = Shell.instance
            ident = threading.get_ident()
            panid = shell.interpreters[ident].console_id            
        return ConsoleGuiProxy._gui_set_console_mode(mode, panid)


    def show_me(self):
        shell = Shell.instance
        this_panid = shell.this_interpreter().console_id  
        self.show(this_panid)

    
    @StaticGuiCall    
    def show(panid):
        console = gui.qapp.panels['console'][panid]
        console.show_me()


    @StaticGuiCall       
    def _gui_set_console_mode(mode='input', panid=None):
        console = gui.qapp.panels['console'][panid]  
        old_mode = console.stdio.stdInputPanel.mode      
        console.set_mode(mode) 
        return old_mode


    @StaticGuiCall       
    def release_side_thread(panid):
        task = gui.qapp.panels['console'][panid].task
        task.release_control()


    def execute_code(self, code_string, panid=None):
        shell = Shell.instance
        this_panid = shell.this_interpreter().console_id        
        if panid is None or this_panid == panid:
            exec(code_string, shell.wsdict)
        else:
            ConsoleGuiProxy._gui_execute_code(code_string, panid)            


    @StaticGuiCall
    def _gui_execute_code(code_string=None, panid=None):
        """
        Execute code in ANOTHER console        
        """
        console = gui.qapp.panels.select_or_new('console', panid) 
        console.stdio.stdInputPanel.execute_commands(code_string)        


    def execute_file(self, filepath, panid=None):
        shell = Shell.instance
        this_panid = shell.this_interpreter().console_id        
        if panid is None or this_panid == panid:
            shell.execfile(filepath, shell.wsdict)
        else:
            ConsoleGuiProxy._gui_execute_file(filepath)


    @StaticGuiCall
    def _gui_execute_file(filepath, panid=None):
        """
        Execute a file in ANOTHER console
        """
        console = gui.qapp.panels.select_or_new('console', panid)
        console.task.send_command(f"_ = shell.execfilews(r'{filepath}')")
        
    @StaticGuiCall        
    def sync_paths(source_panid=0, target_panid=1):
        """
        Syncing sys.path and the live paths from source to target.
        """
        from ...core.shellmod import Shell
        
        source_task = gui.qapp.panels['console'][source_panid].task
        target_task = gui.qapp.panels['console'][target_panid].task
        
        sys_paths = source_task.call_func(Shell.get_sys_paths, wait=True)  
        target_task.call_func(Shell.add_sys_paths, args=(sys_paths,))
        
        live_paths = source_task.call_func(Shell.get_live_paths, wait=True)
        target_task.call_func(Shell.set_live_paths, args=(live_paths,))
    
    @StaticGuiCall    
    def child(init_caller, *args):
        panel = gui.qapp.panels.select_or_new('console', None, 'child')
        #panel.window().toggleStatOnTop(True)
        panel.task.wait_process_ready()        
                    
        panel.task.call_func(init_caller, args=args, callback=None)           
        return panel.panid



import numpy as np
from .base import RecordTable

def compare(table1, table2, value_column_name, index_column_name=None):

    if index_column_name is None:
        valcol1 = table1.sa[value_column_name]
        valcol2 = table2.sa[value_column_name]

        difflocs = np.where(valcol1 != valcol2)

        difftable = RecordTable()
        difftable.add_column('value1', data=valcol1[difflocs])
        difftable.add_column('value2', data=valcol2[difflocs])

    else:
        ind1 = table1.sa[index_column_name]
        ind2 = table2.sa[index_column_name]

        valcol1 = table1.colnames.index(value_column_name)
        valcol2 = table2.colnames.index(value_column_name)

        locs1 = dict(zip(ind1, range(len(ind1))))
        locs2 = dict(zip(ind2, range(len(ind2))))

        indices = sorted(set(ind1).union(set(ind2)))

        difftable = RecordTable(['index', 'value1', 'value2'])

        for ind in indices:
            loc1 = locs1.get(ind, None)
            loc2 = locs2.get(ind, None)
            rec1 = None if loc1 is None else table1[loc1]
            rec2 = None if loc2 is None else table2[loc2]
            if rec1[valcol1] != rec2[valcol2]:
                difftable.add_row((ind, rec1[valcol1], rec2[valcol2]))

    return difftable



import sys
import os
import threading
import multiprocessing
import builtins
import logging
import traceback
import subprocess
import pprint
import shlex
import json
import psutil
import inspect
import shlex
from pathlib import Path

from . import stdinout
from .interpreter import QueueInterpreter
from .stdinout import ProcessStdInput
from .gui_proxy import gui
from .conf import config
from .history import LogDir
   
from ..utils.names import DictStruct 
from ..rectable import RecordTable
from ..live import use, manager

if config['console']['completer'] == 'native':
    from rlcompleter import Completer
else:
    from ..live.completer import Completer

here = Path(__file__).absolute().parent
logger = logging.getLogger(__name__) 
  
class Shell(object):
    instance = None
    
    def __init__(self, workspace=None):
        self.wsdict = dict() if workspace is None else workspace
        self.ws = DictStruct(self.wsdict)
             
        Shell.instance = self
        self.wsdict['shell'] = self        
        self.wsdict['gui'] = gui
        self.wsdict['use'] = use     
        self.wsdict['__name__'] = '__main__'        
        
        self.redirect_stdout()            
        self.redirect_input()
        
        self.comp = Completer(self.wsdict)
        self.interpreters = dict()
        self.logdir = LogDir(config['path_log'])    
        
    def redirect_stdout(self):
        if not config['debug']['skip_main_stdout_redirect']:
            current_stdout = sys.stdout
            sys.stdout = self.stdout = stdinout.StdOutRouter()
            sys.stdout.backup_stream = current_stdout
            stdinout.enable_ghstream_handler()
            
        if not config['debug']['skip_main_stderr_redirect']:
            current_stderr = sys.stderr
            sys.stderr = self.stderr = stdinout.StdErrRouter()
            sys.stderr.backup_stream = current_stderr
            
    def get_watcher_ports(self):
        lock_files = self.logdir.get_active_lock_files()
        lock_files = sorted(lock_files, key=os.path.getmtime)
        ports = []
        
        for lock_file in lock_files:
            info_file = lock_file.parent / 'cmdserver.json'
            if not info_file.exists(): continue
            with open(str(info_file), 'r') as fp:
                content = json.load(fp)
            ports.append(content['port'])

        return ports

    @property
    def _qapp(self):
        from qtpy.QtWidgets import QApplication
        return QApplication.instance()
            
    def redirect_input(self):        
        #ProcessStdInput is missing some function
        #to be a good overwrite of sys.stdin. But it is not needed to overwrite sys.stdin
        #Only, overwriting builtins.input is good enough
        #(original input function doesn't seem to be compatible with ProcessStdInput)
        self.stdin = ProcessStdInput()                   
        self.__input__ = builtins.input           
        builtins.input = self.input     
        sys.displayhook = sys.__displayhook__

    def restore_stdout(self):
        sys.stdout = sys.__stdout__
        sys.stderr = sys.__stderr__
        #The following is used to customize the repr functionaility
        #It also stores to _
        sys.displayhook = sys.__displayhook__ 
        
    def restore_input(self):
        builtins.input = self.__input__
        
    def info(self):
        """
        Print info about the current process
        """
        prc = psutil.Process(os.getpid())
        pyprc = multiprocessing.current_process()
        print(f'name: {prc.name()}')
        print(f'pname: {pyprc.name}')
        print(f'exe: {prc.exe()}')
        print(f'argv: {sys.argv}')
        print(f'cwd: {prc.cwd()}')  
        print(f'pid: {prc.pid}')
        print(f'ppid: {prc.ppid()}')              
        print(f'mem rss: {prc.memory_info().rss}')
        print(f'#threads: {prc.num_threads()}')
        this_thread = threading.currentThread()
        print(f'thread name: {this_thread.name}')
        print(f'tid: {this_thread.ident}')
        
    def this_interpreter(self):
        """
        Return the interpreter related to this thread.
        """
        tid = threading.get_ident()
        #TO DO, support for other threads which are not in a console ?
        #       what about the stdout of these threads
        return self.interpreters.get(tid, None)
        
    def edit_file(self, filename, lineno=0):
        """
        Edit the file with an external editor at a certain lineno.        
        """
        editorExecutable = Path(config['texteditor'])
        
        if not editorExecutable.exists():
            logger.error(f"{editorExecutable} doesn't exist")
            return
        
        if editorExecutable.name == 'notepad++.exe':
            # supply line number argument
            os.spawnl(os.P_NOWAIT, editorExecutable, '"' + str(editorExecutable) + '"', '-n%d'%lineno, '"{0}"'.format(filename))
        else:
            os.spawnl(os.P_NOWAIT, editorExecutable, '"' + str(editorExecutable) + '"', '"{0}"'.format(filename))          
            
    def edit_dbase(self, filename):
        """
        Edit a SQLite file with an external editor.
        """
        executable = config['dbbrowser']
        os.spawnl(os.P_NOWAIT, executable, '"' + executable + '"', '"{0}"'.format(filename))        
        
    def edit(self, object):
        """
        Edit source file of object with external editor
        
        :param objects: object from which the source code will be opened in an editor
        :return: None
        """
        object_type_name = type(object).__name__
        
        if object_type_name == 'ImageDataManager':
            self.edit_dbase(object.fullpath)
            return
        
        (filename, lineno) = self.getcodefile(object)
        
        if not filename is None:
            logger.info('opening "%s" with editor at line %d' % (filename, lineno))
            self.edit_file(filename, lineno)
            
        else:
            logger.warn('Could not find the file for object')
                
    def getcodefile(self, object):
        """
        Find out which source or compiled file an object was defined in.
        """
        
        fi = None
        lineno = 0
                            
        if hasattr(object, 'ls_code'):
            #it is a script callable
            code = object.func.__code__
            fi = code.co_filename
            lineno = code.co_firstlineno
            
        elif hasattr(object, '__func__'):
            code = object.__func__.__code__
            fi = code.co_filename
            lineno = code.co_firstlineno            
            
        else:
            if hasattr(object, '__wrapped__'):
                object = object.__wrapped__
            
            try:
                fi = inspect.getsourcefile(object)
            except:
                fi = None
                
            if fi is None:
                fi = inspect.getsourcefile(type(object))                    
            
            if hasattr(object, '__code__'):
                try:
                    lineno = object.__code__.co_firstlineno
                except:
                    lineno = 1
            else:
                try:
                    lineno = inspect.getlineno(object)
                except AttributeError:
                    lineno = 1
                
        return (fi, lineno)                
                    
    @staticmethod
    def new_interactive_thread(cqs, guiproxy=None, client=True):
        shell = Shell.instance
        
        if client and type(cqs).__name__ == 'ZmqQueues':
            cqs.setup_as_client()
        
        thread = threading.Thread(target=QueueInterpreter.create_and_interact, args=(shell, cqs, guiproxy), name='Interact',daemon=True)
        thread.start()
        return (thread.name, thread.ident)
        
    def popen(self, commands, shell=True, stdin=True):        
        
        #https://eli.thegreenplace.net/2017/interacting-with-a-long-running-child-process-in-python/
        
        if isinstance(commands, str):
            commands = shlex.split(commands)
    
        def output_reader(proc):
            for line in iter(proc.stdout.readline, b''):
                sys.stdout.write(line.decode('utf-8'))
                sys.stdout.flush()
                
            print(f'{commands[0]}: stdout ended')
            #sys.stdout.redirects.pop(threading.get_ident())
            
        def stderr_reader(proc):
            for line in iter(lambda: proc.stderr.read(1), b''):
                sys.stderr.write(line.decode('utf-8'))
                sys.stderr.flush()
                
            print(f'{commands[0]}: stderr ended')
            #sys.stderr.redirects.pop(threading.get_ident())            
        
        if stdin:
            process = subprocess.Popen(commands, stdin=subprocess.PIPE,
                stdout= subprocess.PIPE, stderr= subprocess.PIPE, shell=shell)
        else:
            process = subprocess.Popen(commands, 
                stdout= subprocess.PIPE, stderr= subprocess.PIPE, shell=shell)
            
        stdout_thread = threading.Thread(target=output_reader, args=(process,))
        stdout_thread.start()
        
        stderr_thread = threading.Thread(target=stderr_reader, args=(process,))
        stderr_thread.start()
        
        sys.stdout.copy_to_thread(stdout_thread.ident)
        sys.stderr.copy_to_thread(stderr_thread.ident)
        
        while True:
           cmd = input('')
           if cmd == '': break
           process.stdin.write(cmd.encode() + '\n'.encode())
           process.stdin.flush()      

        process.stdin.close()        
        process.terminate()
                
    # def ipython(self):
        # from IPython.terminal.embed import InteractiveShellEmbed
        # ishell = InteractiveShellEmbed()
        # stdout_back = sys.stdout
        # ishell.interact()
        # sys.stdout = stdout_back
        
    # def jupyter(self, rundir=''):
        # JUPYTER_DATA_DIR = str(here.parent / 'external' / 'jupyter' / 'data')
        # logger.info(f'JUPYTER_DATA_DIR: {JUPYTER_DATA_DIR}')            
        # os.environ['JUPYTER_DATA_DIR'] = JUPYTER_DATA_DIR
        # os.system(f'start {sys.executable} -m jupyterlab --notebook-dir="{rundir}"')
        
    def pprint(self, var):
        """
        Do a `pretty print <https://docs.python.org/3.8/library/pprint.html>`_ of var.
        The user can also use `var!!` to call this function.               
        """
        pprint.pprint(var)
        
    def magic(self, cmd):
        """
        Magic commands like in IPython.
        """
        cmd, *args = shlex.split(cmd)
        if cmd == 'cd':
            os.chdir(args[0])            
            
        elif cmd == 'pwd':
            print(Path('.').resolve())
            
        elif cmd in ['ls', 'dir']:
            self.popen('dir')
            
        elif cmd == 'tb':
            traceback.print_last()
            
        elif cmd == 'info':
            self.info()
            
        elif cmd == 'who':            
            tbl = RecordTable(['name', 'repr', 'str'])
            
            for key in list(self.wsdict.keys()):
                if key.startswith('_'): continue
                obj = self.wsdict[key]
                tbl.add_row((key, repr(obj), str(obj)))
                
            print(tbl)            
        
    def start_in_this_thread(self, cqs):      
        QueueInterpreter.create_and_interact(self, cqs, None)      
        
    @staticmethod
    def get_completer_data(text, max=1000, wild=False):
        shell = Shell.instance
        
        items = []
        for state in range(max):
            if shell.comp.complete.__func__.__code__.co_argcount == 3:
                #The Python Original rlcompleter
                item = shell.comp.complete(text, state)
            else:
                item = shell.comp.complete(text, state, wild)
            if item is None:
                break
            items.append(item)
        return items        
        
    def input(self, message=''):
        ident = threading.get_ident()                
        if ident in self.interpreters.keys():
            if gui.is_main():
                return gui.inputdlg(message)
            else:
                prior_console_mode = gui.console.set_mode('input')
                print(message, end='')
                mode, args, callback = self.stdin.read()
                gui.console.set_mode(prior_console_mode)
                return args[0]
        else:
            return self.__input__(message)
            
    def execfile(self, filepath, globals=None, locals=None):
        """
        Execute a Python file
        """
        filepath = str(Path(filepath).absolute())
        source = open(filepath, 'r').read()
        code = compile(source, filepath, 'exec')
        exec(code, globals, locals)
        
    def execfilews(self, filepath, wsname='__execfilews__'):
        """
        Execute a Python file in a new workspace.
        Place the workspace in shell
        """    
        filepath = str(Path(filepath).absolute())
        ws = dict()
        ws['__file__'] = filepath
        self.execfile(filepath, ws)
        self.wsdict[wsname] = ws
        return ws
        
    @staticmethod        
    def get_sys_paths(customs_only=True):
        """
        List the sys.paths.
        
        :param bool customs_only: List only paths not part of the python.exe directory.
        """
        base = Path(sys.executable).parent
        custom_paths = []
        for path in sys.path:
            path = Path(path)
            if customs_only:
                if base in path.parents:
                    continue
                elif base == path:      
                    continue  
                elif path == Path('.'):
                    continue                
            custom_paths.append(str(path))
        return custom_paths
        
    @staticmethod
    def set_sys_paths(new_sys_paths):
        """
        Set new content for sys.path.
        """
        sys.path.clear()
        sys.path.extend(new_sys_paths)
        print(f'The new paths are {sys.path}')
        
    @staticmethod
    def add_sys_paths(sys_paths):
        for path in sys_paths:
            if not path in sys.path:
                sys.path.append(path)
        
    @staticmethod
    def get_live_paths():
        #from ..live import manager
        return manager.path
        
    @staticmethod
    def set_live_paths(new_live_paths):
        #from ..live import manager     
        manager.path.clear()
        manager.path.extend(new_live_paths)
        print(f'The new live paths are {manager.path}')
        
        
    



import os
import sys
import queue
import threading
import collections
import socket
import json
import multiprocessing
import _multiprocessing

import zmq
zmq_context = zmq.Context()

from .conf import config

if sys.platform == 'win32':
    from .comm_nonduplex import NonDuplexQueue   

sentinel = object()

class CommQueues(object):
    def __init__(self, QueueCls, process=False):
        self.host = 'localhost'
        self.cmd_queue = QueueCls()
        self.return_queue = QueueCls()

        self.stdin_queue = QueueCls()
        self.stdout_queue = QueueCls()

        if process:
            self.gui_call_queue = QueueCls()
            self.gui_return_queue = QueueCls()
        else:
            self.gui_call_queue = None
            self.gui_return_queue = None

    def clear(self):
        while not self.cmd_queue.empty():
            self.cmd_queue.get()

        while not self.stdout_queue.empty():
            self.stdout_queue.get()

        while not self.return_queue.empty():
            self.return_queue.get()

        if not self.gui_call_queue is None:
            while not self.gui_call_queue.empty():
                self.gui_call_queue.get()

        if not self.gui_return_queue is None:
            while not self.gui_return_queue.empty():
                self.gui_return_queue.get()

    def close(self):
        for q in [self.cmd_queue, self.stdout_queue, self.return_queue,
                  self.gui_call_queue, self.gui_return_queue]:
            if isinstance(q, NonDuplexQueue):
                q.close()


class ZmqQueue(object):

    def __init__(self, port=None):
        self.port = port

    def setup_as_server(self):
        self.socket = zmq_context.socket(zmq.PAIR)
        self.port = self.socket.bind_to_random_port('tcp://*',
            min_port=config['zmq_queue_min_port'],
            max_port=config['zmq_queue_max_port'],
            max_tries=100)

    def setup_as_client(self, host='localhost'):
        self.socket = zmq_context.socket(zmq.PAIR)
        self.socket.connect(f"tcp://{host}:{self.port}")

    def __getstate__(self):
        state = dict()
        state['port'] = self.port
        return state

    def __setstate__(self, state):
        self.port = state['port']

    def put(self, data):
        self.socket.send_pyobj(data, flags=zmq.NOBLOCK)

    def get(self, timeout=None):
        if not timeout is None:
            event = self.socket.poll(timeout*1000)
            if event == 0:
                raise queue.Empty()
        return self.socket.recv_pyobj()

    def empty(self):
        return self.socket.poll(0) == 0


class ZmqQueues(object):
    ports = {
        'cmd': None,
        'stdin': None,
        'stdout': None,
        'return': None,
        'gui_call': None,
        'gui_return': None,
        }

    def __init__(self, ports=None):
        if ports is None:
            ports = ZmqQueues.ports

        for queue_name, port in ports.items():
            self.__dict__[f'{queue_name}_queue'] = ZmqQueue(port=port)

    def setup_host(self):
        self.hostname = socket.gethostname()
        hostname_ex, aliaslist, ipaddrlist = socket.gethostbyname_ex(self.hostname)
        self.hostname_ex = hostname_ex
        self.aliaslist = aliaslist
        self.ipaddrlist = ipaddrlist

    @classmethod
    def from_json(Cls, json_string):
        d0 = json.loads(json_string)
        d1 = d0['channel']

        ports = dict()
        ports['cmd'] = d1['cmd']
        ports['stdin'] = d1['stdin']
        ports['stdout'] = d1['stdout']
        ports['return'] = d1['return']
        ports['gui_call'] = d1['gui_call']
        ports['gui_return'] = d1['gui_return']

        instance = Cls(ports)

        instance.hostname = d1['hostname']
        instance.hostname_ex = d1['hostname_ex']
        instance.aliaslist = d1['aliaslist']
        instance.ipaddrlist = d1['ipaddrlist']

        return instance

    def to_json(self):
        d0 = dict()
        d1 = d0['channel'] = dict()
        d1['method'] = 'zmq'

        d1['hostname'] = self.hostname
        d1['hostname_ex'] = self.hostname_ex
        d1['aliaslist'] = self.aliaslist
        d1['ipaddrlist'] = self.ipaddrlist

        d1['cmd'] = self.cmd_queue.port
        d1['stdin'] = self.stdin_queue.port
        d1['stdout'] = self.stdout_queue.port
        d1['return'] = self.return_queue.port
        d1['gui_call'] = self.gui_call_queue.port
        d1['gui_return'] = self.gui_return_queue.port
        return json.dumps(d0)

    def setup_as_server(self):
        self.setup_host()
        self.cmd_queue.setup_as_server()
        self.stdin_queue.setup_as_server()
        self.stdout_queue.setup_as_server()
        self.return_queue.setup_as_server()
        self.gui_call_queue.setup_as_server()
        self.gui_return_queue.setup_as_server()

    def setup_as_client(self, host=None):
        if host is None:
            self.host = 'localhost'
        else:
            self.host = host

        self.cmd_queue.setup_as_client(self.host)
        self.stdin_queue.setup_as_client(self.host)
        self.stdout_queue.setup_as_client(self.host)
        self.return_queue.setup_as_client(self.host)
        self.gui_call_queue.setup_as_client(self.host)
        self.gui_return_queue.setup_as_client(self.host)



"""
Small utils directly related to QT
"""
import collections
from pathlib import Path
from qtpy import QtWidgets

from .. import config

respath = Path(config['respath'])

def relax_menu_text(text):
    """Remove any &, lower case, stop at first \\t"""
    tmp = text.replace('&', '').lower()
    tabind = tmp.find('\t')
    if tabind > 0:
        tmp = tmp[:tabind]
    return tmp.strip()

def relax_menu_trace(menutrace):
    """relax_menu_text on every string of menutrace"""
    return tuple(relax_menu_text(item) for item in menutrace)

def getMenuTrace(menu):
    """Return the current menu location as a list of strings"""
    menutrace = [menu.title().strip('&')]
    scan_action = menu.menuAction()
    while not scan_action is None:
        aw = scan_action.associatedWidgets()
        if len(aw) > 0 and isinstance(aw[0], QtWidgets.QMenu):
            menu = aw[0]
            menutrace.append(menu.title().strip('&'))
            scan_action = menu.menuAction()
        else:
            break

    return menutrace[::-1]

def getMenuAction(menubar, menutrace):
    """Locate the action in a menubar"""
    actions = menubar.actions()

    menutrace = relax_menu_trace(menutrace)

    for check_action_name in menutrace:
        for action in actions:
            action_name = action.text()
            action_name = relax_menu_text(action_name)

            if action_name == check_action_name:
                menu = action.menu()
                if not menu is None:
                    #Cause showEvent which is sometimes used to refresh the menu content
                    menu.show()
                    actions = action.menu().actions()
                    menu.hide()
                else:
                    actions = []
                break

        else:
            raise KeyError(f'Action part "{check_action_name}" not found')

    return action
    
class ActionArguments(object):
    def __init__(self, qwidget):
        self.qwidget = qwidget
        self.args = collections.OrderedDict()
        
    def __setitem__(self, key, value):
        self.args[key] = value
        
    def __getitem__(self, key):
        return self.args[key]

    def isNotSet(self):
        return self.qwidget.sender().data() is None
        
    def __enter__(self):        
        return self
        
    def __exit__(self, *args, **kwargs):
        arguments = self.qwidget.sender().data() or {}
        self.args.update(dict(zip(self.args.keys(), arguments.get('args', []))))
        self.args.update(arguments.get('kwargs', {}))



import platform
import numpy as np
import logging

logger = logging.getLogger(__name__)

from ...core.gui_proxy import GuiProxyBase, StaticGuiCall
from ... import gui, config
from ...utils.shared import SharedArray
        
class ImageGuiProxy(GuiProxyBase):    
    category = 'image'
    opens_with = ['.tif', '.png', '.gif']
    
    def __init__(self):
        pass
        
    def attach(self, gui):
        gui.img = self   
        
        # Properties are defined on gui 
        
        gui.set_roi = self.set_roi
        gui.set_roi_slices = self.set_roi_slices
        gui.get_roi = self.get_roi
        gui.get_roi_slices = self.get_roi_slices
        gui.jump_to = self.jump_to
        gui.zoom_fit = self.zoom_fit
        gui.zoom_full = self.zoom_full
        gui.zoom_region = self.zoom_region
        gui.get_clipboard_image = self.get_clipboard_image
        
        return 'img'
        
    @StaticGuiCall
    def new(cmap=None, viewtype='image-profile', title=None):
        panel = GuiProxyBase._new('image', viewtype)
        
        if not cmap is None:
            panel.colormap = cmap        
            
        if not title is None:
            panel.long_title = title
            
        return panel.panid
    
    @StaticGuiCall    
    def open(filepath):
        panel = gui.qapp.panels.selected('image')
        if panel is None:
            panel = gui.qapp.panels.select_or_new('image', defaulttype = 'image-profile')
        panel.openImage(filepath)
        window = panel.get_container().parent()
        window.raise_()
        gui.qapp.processEvents()
        return panel.panid
        
    @StaticGuiCall    
    def get_clipboard_image():
        from ...utils.imconvert import qimage_to_ndarray
        
        cb = gui.qapp.clipboard()
        md = cb.mimeData()
        qimg = md.imageData()
        if qimg is None:
            logger.error('No image data on clipboard')
            return
        arr = qimage_to_ndarray(qimg)
        return arr              
        
    @StaticGuiCall
    def select(panid=-1):
        """
        Select or create an image panel with id image_id or auto id
        """        
        panel = gui.qapp.panels.select_or_new('image', panid)
        return panel.panid
    
    def show(self, array, cmap=None):   
        """
        Show array in the image viewer.
        
        :param ndarray array: 
        :param str cmap: 'grey', 'jet' or 'turbo'
        """
        
        if config['image']['queue_array_shared_mem']:
            sharr = SharedArray(array.shape, array.dtype)
            sharr[:] = array          
            return ImageGuiProxy.show_array(sharr, cmap)
                
        else:
            return ImageGuiProxy.show_array(array, cmap)

    @StaticGuiCall
    def show_array(array=None, cmap=None):
        panel = gui.qapp.panels.selected('image')

        if not cmap is None:
            panel.colormap = cmap        

        panel.show_array(array)
        return panel.panid
    
    @StaticGuiCall    
    def show_mask(array=None, composition='sourceover'):
        panel = gui.qapp.panels.selected('image')
        panel.imviewer.imgdata.set_mask(array, composition)
        panel.imviewer.refresh()
       
    def refresh(self):
        ImageGuiProxy.show_array()        

    @StaticGuiCall
    def cmap(cmap):
        ImageGuiProxy.show_array(None, cmap)
        
    @property    
    def vs(self):
        shared = config['image']['queue_array_shared_mem']
        
        if shared:
            return self.get_image_view_source(True).ndarray
        else:
            return self.get_image_view_source(False)
        
    @property    
    def vr(self):
        return self.get_image_view_region()

    @property    
    def buff(self):
        return self.get_image_view_buffer()         
        
    @StaticGuiCall
    def get_image_view_source(shared=True):            
        panel = gui.qapp.panels.selected('image')
        if panel is None: return
        
        if shared:
            return panel.sharray
        else:
            return panel.sharray.ndarray
        
    def get_image_view_region(self):
        slices =  self.get_roi_slices()
        if slices is None: return None
        vs = self.get_image_view_source()
        if vs is None: return
        return vs[slices] 
        
    @StaticGuiCall
    def get_image_view_buffer():
        panel = gui.qapp.panels.selected('image')
        if panel is None: return None
        return panel.imviewer.imgdata.array8bit
        
    @StaticGuiCall
    def repaint():
        panel = gui.qapp.panels.selected('image')
        panel.imviewer.refresh()
        
    @StaticGuiCall
    def set_range(black, white):
        panel = gui.qapp.panels.selected('image')
        panel.changeBlackWhite(black, white)        
        
    @staticmethod
    def close_all():
        while ImageGuiProxy.close():
            pass
        
    @StaticGuiCall
    def zoom_fit():
        """
        Zoom the image to fitting the image viewer area.
        Snap on the default zooming values.
        """
        panel = gui.qapp.panels.selected('image')
        panel.zoomFit()

    @StaticGuiCall
    def zoom_full():
        """
        Zoom the image to fully fitting the image viewer area.
        """
        panel = gui.qapp.panels.selected('image')
        panel.zoomFull()  

    @StaticGuiCall
    def zoom_region(x, y, width, height):
        """
        Zoom the image to a certain region.
        """
        panel = gui.qapp.panels.selected('image')
        panel.zoomToRegion(x, y, width, height) 

    @StaticGuiCall
    def jump_to(x, y):
        """
        Select a certain pixel and zoom to it
        """ 
        panel = gui.qapp.panels.selected('image')
        panel.jumpTo(x, y)        
        
    @StaticGuiCall
    def set_roi_slices(slices, yonfirst=True):
        """
        Set the region of interest on the current viewport.        
        
        :param tuple slices: Tuple of slices accross the dimensions.
        :param bool yonfirst: True if first slice is the y direction.
        """        
        panel = gui.qapp.panels.selected('image')
        roi = panel.imviewer.roi
        selroi = panel.imviewer.imgdata.selroi        
        
        if yonfirst:
            selroi.xr.setfromslice(slices[1])
            selroi.yr.setfromslice(slices[0])        
        else:
            selroi.xr.setfromslice(slices[0])
            selroi.yr.setfromslice(slices[1])
        
        roi.clip()
        roi.show()
        roi.roiChanged.emit()    
        
    @StaticGuiCall
    def set_roi(x0, y0, width=1, height=1):
        """
        Set the region of interest on the current viewport.
        
        :param int x0: first column of the roi
        :param int y0: first row of the roi
        :param int width: width of the roi
        :param int height: height of the roi        
        """    
        panel = gui.qapp.panels.selected('image')
        roi = panel.imviewer.roi
        roi.setStartEndPoints(x0, y0, x0 + width - 1, y0 + height - 1)        
        roi.show()
        roi.roiChanged.emit()      

    @StaticGuiCall
    def get_roi_slices():
        """
        Get the current region of interest as a tupple of slice objects
        """
        panel = gui.qapp.panels.selected('image')
        if panel is None: return
        return panel.imviewer.imgdata.selroi.getslices()
        
    @StaticGuiCall
    def get_roi():
        """
        Get the region of interest of the current viewport as a tuple of integers.
        
        :return tuple(int, int, int, int): x0, y0, width, height.
        """
        slices = ImageGuiProxy.get_roi_slices()
        x0 = slices[1].start
        width = slices[1].stop - x0
        y0 = slices[0].start
        height = slices[0].stop - y0
        return x0, y0, width, height   

    @property
    def vr(self):
        slices =  self.get_roi_slices()
        return self.vs[slices]        
        
    @staticmethod
    def mirror_x():
        print('mirror_x started')
        arr = gui.vs.copy()
        arr = arr[:,::-1]
        gui.img.new()
        gui.img.show(arr)        
        print('mirror_x ended')
        
    @staticmethod
    def high_pass_current_image():
        logger.error('Not implemented')    
        
    @staticmethod
    def get_distance():
        import math
        
        print('Select first point: ', end='')
        p1 = ImageGuiProxy.get_selected_pixel()
        print(p1)
        print('Select second point: ', end='')
        p2 = ImageGuiProxy.get_selected_pixel()
        print(p2)
        delta = (p2[0] - p1[0], p2[1] - p1[1])
        print(f'delta xy: {delta}')
        print(f'delta  r: {(delta[0]**2 + delta[1]**2) ** 0.5:.5g}')
        if delta[0] == 0:
            print(f'angle  r: {90}')
        else:
            print(f'angle  r: {math.atan(delta[1]/delta[0]) * 180 / 3.141592:.5g}')
            
    @staticmethod
    def get_selected_pixel():        
        ImageGuiProxy._push_selected_pixel_queue(True)
        pixel = input()
        return eval(pixel)            
    
    @StaticGuiCall      
    def _push_selected_pixel_queue(enable=True):
        panel = gui.qapp.panels.selected('image')
        panel.imviewer.push_selected_pixel = enable


import pathlib
import collections
import queue
import threading
import math

from qtpy import QtGui, QtCore
from qtpy.QtGui import QImage

import numpy as np

from ... import gui, config

from ...utils.shared import SharedArray
from ...utils import imconvert

from .dimensions import DimRanges
from . import fasthist

here = pathlib.Path(__file__).absolute().parent

try:
    from .numba_func import map_values_mono, map_values_rgbswap, map_values_rgb
    has_numba = True
    
except:
    has_numba = False

#https://doc.qt.io/qtforpython-5.12/PySide2/QtGui/QPainter.html#composition-modes
COMPMODE = dict()    
COMPMODE['sourceover'] = QtGui.QPainter.CompositionMode_SourceOver    
COMPMODE['plus'] = QtGui.QPainter.CompositionMode_Plus   
COMPMODE['multiply'] = QtGui.QPainter.CompositionMode_Multiply   
COMPMODE['screen'] = QtGui.QPainter.CompositionMode_Screen  
COMPMODE['overlay'] = QtGui.QPainter.CompositionMode_Overlay  
COMPMODE['darken'] = QtGui.QPainter.CompositionMode_Darken  
COMPMODE['lighten'] = QtGui.QPainter.CompositionMode_Lighten 
    
class SelectRoi(DimRanges):

    """
    Selection widget range data.
    """

    def __init__(self, height, width, update_statistics_func=None):
        DimRanges.__init__(self, (height, width))
        self.update_statistics_func = update_statistics_func
        
    @property
    def yr(self):
        return self.rngs[0]
    
    @property    
    def xr(self):
        return self.rngs[1]

    def ensure_rising(self):
        for rng in self.rngs:
            if rng.start > rng.stop:
                rng.start, rng.stop = rng.stop, rng.start

    def copy(self):
        s = SelectRoi(self.rngs[0].maxstop, self.rngs[1].maxstop)
        s.inherite(self)
        return s
        
    def update_statistics(self):
        if not self.update_statistics_func is None:
            self.update_statistics_func()
        
class ImageStatistics(object):

    def __init__(self):
        self._cache = dict()
        self.arr2d = None        
        
    def attach_arr2d(self, arr2d):
        self.arr2d = arr2d
        self.clear()
        
    @property
    def dtype(self):
        return self.arr2d.dtype        
        
    def clear(self):
        self._cache.clear()
        
    def step_for_bins(self, bins):
        if self.dtype in ['float16', 'float32', 'float64']:
            return math.ceil(65536 / bins) 
            
        if len(self._cache.keys()) == 0:
            self.calc_histogram()
            
        hist1 = self._cache['hist']  
        
        return math.ceil(len(hist1) / bins)
        
    def histogram(self, step=1):
        if len(self._cache.keys()) == 0:
            self.calc_histogram()
        
        hist1 = self._cache['hist']      
            
        if step > 1:
            bins = len(hist1) // step
            left = len(hist1) % step
            tmp = hist1[:step*bins]            
            if left > 0:
                hist = np.r_[tmp.reshape(bins, step).sum(1), hist1[step*bins:].sum()]                
            else:
                hist = tmp.reshape(bins, step).sum(1)
            return hist
        else:
            return hist1       
        
    def starts(self, step=1):
        if len(self._cache.keys()) == 0:
            self.calc_histogram()
            
        starts1 = self._cache['starts'] 

        if step > 1:
            return starts1[::step]
        else:        
            return starts1      
    
    def calc_histogram(self, bins=None, step=None):        
        if self.dtype in ['int8', 'uint8', 'int16', 'uint16']:
            hist, starts, stepsize = fasthist.hist16bit(self.arr2d, bins=None, step=1, use_numba=True)
            
        elif self.dtype in ['float16', 'float32', 'float64']:
            hist, starts, stepsize = fasthist.histfloat(self.arr2d, bins=65536, step=None, pow2snap=False, use_numba=True)
            
        self._cache['hist'] = hist
        self._cache['starts'] = starts
        self._cache['stepsize'] = stepsize            
    
    @property    
    def bins(self):
        return len(self.starts())
        
    def stepsize(self, step):
        return self._cache['stepsize'] * step
        
    def n(self):
        return self.arr2d.shape[0] * self.arr2d.shape[1]
        
    def sum(self):
        return (self.histogram() * self.starts()).sum()
        
    def mean(self):
        return self.sum() / self.n()

    def sumsq(self):
        return (self.histogram() * self.starts()**2).sum()
        
    def min(self):
        non_zeros_indices = np.argwhere(self.histogram() > 0)
        min_index = non_zeros_indices[0][0]
        max_index = non_zeros_indices[-1][0]
        return self.starts()[min_index]
        
    def max(self):
        non_zeros_indices = np.argwhere(self.histogram() > 0)
        max_index = non_zeros_indices[-1][0]
        return self.starts()[max_index]     
        
    def std(self):
        n = self.n()
        result = ((self.sumsq() - ((self.sum() * 1.0) ** 2) / n) / (n - 1)) ** 0.5
        return result
        

class ImageData(object):        
    def __init__(self):
        self.qimg = None
        self.map8 = None
        self.sharray = None
        self.imghist = ArrayHistory(config['image'].get("history_size", 500e6))
        
        #arr = np.zeros((192,256),'uint8') + np.arange(256, dtype='uint8').reshape(1,256) 
        arr = np.ones((1,1),'uint8') * 128
        self.selroi = SelectRoi(1, 1, self.update_roi_statistics)
        self.chanstats = dict()
        
        self.show_array(arr)
        self.layers = collections.OrderedDict()
    
    def load_by_qt(self, path):
        self.qimg = QImage(str(path))
        
    def show_array(self, array=None, offset=0, gain=1.0, colormap=None, gamma=1):
        with gui.qapp.waitCursor():
            threadcount = config['image']['threads'] 
            use_numba = config['image']['numba'] and has_numba        
            
        
            if array is None:
                #offset and gain adjust of current viewer
                array = self.sharray
                shape = array.shape
                dtype = array.dtype            
                for stat in self.chanstats.values():
                    stat.clear()
                
            else:
                if not self.sharray is None:
                    self.imghist.push(self.sharray)
                
                if isinstance(array, SharedArray):
                    self.sharray = array            
                    shape = array.shape
                    dtype = array.dtype                           
                    
                else:
                    shape = array.shape
                    dtype = array.dtype
                    self.sharray = SharedArray(shape, dtype)
                    self.sharray[:] = array   

                self.chanstats.clear()
                    
                if len(shape) == 2:
                    self.chanstats['K'] = ImageStatistics()
                    self.chanstats['K'].attach_arr2d(self.sharray.ndarray)
                    self.chanstats['RK'] = ImageStatistics()
                    self.update_roi_statistics()
                else:
                    self.chanstats['R'] = ImageStatistics()
                    self.chanstats['G'] = ImageStatistics()
                    self.chanstats['B'] = ImageStatistics()
                    self.chanstats['R'].attach_arr2d(self.sharray.ndarray[:,:,0])
                    self.chanstats['G'].attach_arr2d(self.sharray.ndarray[:,:,1])
                    self.chanstats['B'].attach_arr2d(self.sharray.ndarray[:,:,2])                   
                    self.chanstats['RR'] = ImageStatistics()
                    self.chanstats['RG'] = ImageStatistics()
                    self.chanstats['RB'] = ImageStatistics()
                    self.update_roi_statistics()
                    
            height, width, *ignore = shape
            self.height, self.width = height, width
                
            if self.selroi.isfullrange():
                self.selroi.xr.maxstop = width
                self.selroi.yr.maxstop = height
                self.selroi.reset()
            else:
                self.selroi.xr.maxstop = width
                self.selroi.yr.maxstop = height
                self.selroi.clip()
                                           
            self.array8bit, self.qimg = imconvert.process_ndarray_to_qimage_8bit(
                self.sharray.ndarray, offset, gain, colormap, refer=True, shared=config["image"].get("qimg_shared_mem", False),
                gamma=gamma)
        
    def get_natural_range(self):
        return imconvert.natural_range(self.sharray.dtype)
        
    def set_mask(self, array=None, composition='sourceover'):
        self.set_layer('mask', array, composition)        
        
    def set_layer(self, name, array=None, composition='sourceover'):
        if array is None:
            if name in self.layers.keys():
                self.layers.pop(name)
            return
            
        assert array.ndim == 2
        assert array.dtype in ['uint8', 'bool']
        
        height, width = array.shape
        
        compmode = COMPMODE[composition.lower()]
            
        qimage = QImage(memoryview(array), width, height, width, QImage.Format_Indexed8)
        qimage.setColorTable(imconvert.make_color_table('mask'))
        self.layers[name] = {'array': array, 'qimage': qimage, 'composition': compmode}                
        
        
    def update_roi_statistics(self):
        slices = self.selroi.getslices()
            
        clr_slices = {'RK': slices,
            'RR': (slices[0], slices[1], 0),
            'RG': (slices[0], slices[1], 1),
            'RB': (slices[0], slices[1], 2)}
            
        for clr, chanstat in self.chanstats.items():  
            if not clr in clr_slices.keys(): continue
            chanstat.attach_arr2d(self.sharray[clr_slices[clr]])

    def update_array8bit_by_slices(self, slices):
        def takemap(source_slice, target_slice):
            self.array8bit[target_slice] = np.take(self.map8, self.sharray.ndarray[source_slice])
        
        threads = []
        for (source_slice,  target_slice) in slices:
            threads.append(threading.Thread(target=takemap, args=(source_slice, target_slice)))
        
        for thread in threads:
            thread.start()
            
        for thread in threads:
            thread.join()   
    
    @property
    def statarr(self):
        return self.sharray.ndarray          
        
    def get_number_of_bytes(self): 
        nbytes = 0
        nbytes += self.sharray.ndarray.nbytes
        nbytes += self.array8bit.nbytes
        return nbytes
        
class ArrayHistory(object):

    def __init__(self, max_size=4):
        self.max_size = max_size 
        self.prior_arrays = []
        self.next_arrays = []
        
    def push(self, array):
        overflow = self.reduce_size_to_max_byte_size(array.size)  
        if overflow < 0: self.prior_arrays.append(array)
        self.next_arrays.clear()        
        
    def reduce_size_to_max_byte_size(self, add_size=0):
        current_size = add_size
        for arr in self.prior_arrays:
            current_size += arr.size
        overflow = current_size - self.max_size
        while overflow > 0 and len(self.prior_arrays) > 0:
            array = self.prior_arrays.pop(0)            
            overflow -= array.size        
        return overflow
        
    def prior(self, current_array):
        array = self.prior_arrays.pop(-1)
        self.next_arrays.append(current_array)
        return array
        
    def next(self, current_array):
        array = self.next_arrays.pop(-1)
        self.prior_arrays.append(current_array)  
        return array
        
    def __len__(self):
        return len(self.prior_arrays) + len(self.next_arrays)
        
    def prior_length(self):
        return len(self.prior_arrays)
        
    def next_length(self):
        return len(self.next_arrays)        
        
    def clear(self):
        self.stack.clear()


from qtpy import QtGui, QtWidgets, QtCore

from ..widgets.thumbs import Thumbs
from ..utils import imconvert

class ColorMapDialog(QtWidgets.QDialog):

    def __init__(self):
        super().__init__()
        colormaps = imconvert.colormaps   
        self.thumbs = Thumbs(self)
        
        for cm_name in colormaps:        
            qimg = imconvert.color_table_preview_qimg(cm_name)
            self.thumbs.addQImage(qimg, cm_name)
            
        self.vbox = QtWidgets.QVBoxLayout(self)
        self.setLayout(self.vbox)
        self.vbox.addWidget(self.thumbs)
        
        self.okBtn = QtWidgets.QPushButton('Ok')
        self.okBtn.clicked.connect(self.ok)
        self.vbox.addWidget(self.okBtn)
        
        self.cm_name = None
        
    def ok(self):
        items = self.thumbs.selectedItems()
        if len(items) > 0:
            self.cm_name = items[0].text()
        else:
            self.cm_name = None
        self.accept()
        
        


import re
import fnmatch

from qtpy import QtCore, QtGui, QtWidgets


class SelectionList(QtWidgets.QListWidget):

    def __init__(self, parent):
        super().__init__(parent=parent)


class FilterList(QtWidgets.QDialog):

    def __init__(self, items, multiple=True, filter='*'):
        super().__init__(None)
        self.initgui(items, multiple, filter)
        
    def initgui(self, items, multiple, filter):
        self.selectionlist = SelectionList(self)
        self.items = items
        
        self.changeSelectStateAllBtn = QtWidgets.QPushButton('Check none', self)
        self.changeSelectStateAllBtn.clicked.connect(self.toggleCheckStateAll)

        self.filter = QtWidgets.QLineEdit(filter, self)
        
        self.useregexpr = QtWidgets.QCheckBox('RegExp', self)
        
        self.addSelectByFilterBtn = QtWidgets.QPushButton('Add', self)
        self.addSelectByFilterBtn.clicked.connect(self.addSelectByFilter)  
        
        self.removeSelectByFilterBtn = QtWidgets.QPushButton('Remove', self)
        self.removeSelectByFilterBtn.clicked.connect(self.removeSelectByFilter)
        
        if not multiple:
            self.changeSelectStateAllBtn.hide()
            self.filter.hide()
            self.useregexpr.hide()            
            self.addSelectByFilterBtn.hide()            
            self.removeSelectByFilterBtn.hide()            
            
        
        self.okBtn = QtWidgets.QPushButton('Ok', self)
        self.okBtn.clicked.connect(self.ok)
        
        self.cancelBtn = QtWidgets.QPushButton('Cancel', self)
        self.cancelBtn.clicked.connect(self.cancel)
        
        layout = QtWidgets.QVBoxLayout()
        layout.addWidget(self.changeSelectStateAllBtn)
        layout.addWidget(self.filter)
        
        hlayout = QtWidgets.QHBoxLayout()
        layout.addLayout(hlayout)
        hlayout.addWidget(self.useregexpr)        
        hlayout.addWidget(self.addSelectByFilterBtn)        
        hlayout.addWidget(self.removeSelectByFilterBtn)        
        
        layout.addWidget(self.selectionlist)
        
        hlayout = QtWidgets.QHBoxLayout()
        layout.addLayout(hlayout)        
        hlayout.addWidget(self.okBtn)
        hlayout.addWidget(self.cancelBtn)
            
        self.setLayout(layout)
        
        for item in self.items:
            list_item = QtWidgets.QListWidgetItem(str(item))
            if multiple:
                list_item.setFlags(QtCore.Qt.ItemIsUserCheckable | QtCore.Qt.ItemIsEnabled)
            else:
                list_item.setFlags(QtCore.Qt.ItemIsSelectable | QtCore.Qt.ItemIsEnabled)
            self.selectionlist.addItem(list_item)            
            
    def checkedItems(self):
        selection = []
        for i in range(self.selectionlist.count()):
            uiitem = self.selectionlist.item(i)
            if uiitem.checkState() == QtCore.Qt.Checked:
                selection.append(self.items[i])
        return selection
        
    def selectedItem(self):
        for i in range(self.selectionlist.count()):
            uiitem = self.selectionlist.item(i)
            if uiitem.isSelected():
                return self.items[i]
        return None
        
    def selectItem(self, item):
        for i in range(self.selectionlist.count()):         
            selection_item = self.selectionlist.item(i)
            if self.items[i] == item:
                selection_item.setSelected(True)
                self.selectionlist.scrollToItem(selection_item)
                return            
        
    def toggleCheckStateAll(self):
        if self.changeSelectStateAllBtn.text() == 'Check all':
            state = True
            self.changeSelectStateAllBtn.setText('Check none')
            self.checkAll(state)
        elif self.changeSelectStateAllBtn.text() == 'Check none':
            state = False
            self.changeSelectStateAllBtn.setText('Check all')
            self.checkAll(state)         

    def addSelectByFilter(self):
        use_regexp = self.useregexpr.isChecked()
        self.checkItemsFilterMatch(self.filter.text(), append=True, remove=False, use_regexp=use_regexp)
            
    def removeSelectByFilter(self):
        use_regexp = self.useregexpr.isChecked()
        self.checkItemsFilterMatch(self.filter.text(), append=False, remove=True, use_regexp=use_regexp)          
            
    def checkItems(self, items_to_select):
        for i in range(self.selectionlist.count()):         
            selection_item = self.selectionlist.item(i)
            if self.items[i] in items_to_select:                            
                selection_item.setCheckState(QtCore.Qt.Checked)                
            else:
                selection_item.setCheckState(QtCore.Qt.Unchecked)
                
    def checkItemsRe(self, pattern, append=True, remove=True):
        self.checkItemsFilterMatch(pattern, append, remove, use_regexp=True)                      
                
    def checkItemsFilterMatch(self, pattern, append=True, remove=True, use_regexp=False):
    
        def matches_fn(item):
            return fnmatch.fnmatch(item, pattern)
            
        def matches_regexp(item):
            return not re.match(pattern, str(item)) is None
    
        if use_regexp:
            matchfunc = matches_regexp
        else:
            matchfunc = matches_fn
            
        for i in range(self.selectionlist.count()):         
            selection_item = self.selectionlist.item(i)
            item = self.items[i]
            
            matches = matchfunc(item)
            #checked = selection_item.getChecked()
            
            if append and remove:
                if matches:
                    selection_item.setCheckState(QtCore.Qt.Checked)
                else:
                    selection_item.setCheckState(QtCore.Qt.Unchecked)
            elif append:
                if matches:
                    selection_item.setCheckState(QtCore.Qt.Checked)
                
            elif remove:
                if matches:
                    selection_item.setCheckState(QtCore.Qt.Unchecked)
            
        
    def checkAll(self, state=True):
        if state:
            newstate = QtCore.Qt.Checked
        else:
            newstate = QtCore.Qt.Unchecked
            
        for i in range(self.selectionlist.count()):
            item = self.selectionlist.item(i)
            item.setCheckState(newstate)
            
    def ok(self):
        self.done(QtWidgets.QDialog.DialogCode.Accepted)
        
    def cancel(self):
        self.done(QtWidgets.QDialog.DialogCode.Rejected)



import ctypes
GW_HWNDNEXT = 2

def get_windows():
    '''Returns windows in z-order (top first)'''
    user32 = ctypes.windll.user32
    lst = []
    top = user32.GetTopWindow(None)
    if not top:
        return lst
    lst.append(top)
    while True:
        next = user32.GetWindow(lst[-1], GW_HWNDNEXT)
        if not next:
            break
        lst.append(next)
    return lst
    
    
def get_z_values(*windows):
    """
    Order these window in z order
    Top window first and lowest z value
    returns a list of [(z, window), ...]
    """
    all_winids_zorder = get_windows()
    zorder = ((all_winids_zorder.index(window.winId()), window) for window in windows if window.winId() in all_winids_zorder)
    return sorted(zorder, key = lambda item: item[0])


import struct

import numpy as np

from qtpy import QtCore, QtGui

Colors = {
    'b': (0,0,255,255),
    'g': (0,255,0,255),
    'r': (255,0,0,255),
    'c': (0,255,255,255),
    'm': (255,0,255,255),
    'y': (255,255,0,255),
    'k': (0,0,0,255),
    'w': (255,255,255,255),
}

def intColor(index, hues=9, values=1, maxValue=255, minValue=150, maxHue=360, minHue=0, sat=255, alpha=255, **kargs):
    """
    Creates a QColor from a single index. Useful for stepping through a predefined list of colors.

    The argument *index* determines which color from the set will be returned. All other arguments determine what the set of predefined colors will be

    Colors are chosen by cycling across hues while varying the value (brightness).
    By default, this selects from a list of 9 hues."""
    hues = int(hues)
    values = int(values)
    ind = int(index) % (hues * values)
    indh = ind % hues
    indv = ind / hues
    if values > 1:
        v = minValue + indv * ((maxValue-minValue) / (values-1))
    else:
        v = maxValue
    h = minHue + (indh * (maxHue-minHue)) / hues

    c = QtGui.QColor()
    c.setHsv(h, sat, v)
    c.setAlpha(alpha)
    return c

def mkColor(*args):
    """
    Convenience function for constructing QColor from a variety of argument types. Accepted arguments are:

    ================ ================================================
     'c'             one of: r, g, b, c, m, y, k, w
     R, G, B, [A]    integers 0-255
     (R, G, B, [A])  tuple of integers 0-255
     float           greyscale, 0.0-1.0
     int             see :func:`intColor() <pyqtgraph.intColor>`
     (int, hues)     see :func:`intColor() <pyqtgraph.intColor>`
     "RGB"           hexadecimal strings; may begin with '#'
     "RGBA"
     "RRGGBB"
     "RRGGBBAA"
     QColor          QColor instance; makes a copy.
    ================ ================================================
    """
    err = 'Not sure how to make a color from "%s"' % str(args)
    if len(args) == 1:
        if isinstance(args[0], QtGui.QColor):
            return QtGui.QColor(args[0])
        elif isinstance(args[0], float):
            r = g = b = int(args[0] * 255)
            a = 255
        #elif isinstance(args[0], basestring):
        elif isinstance(args[0], str):
            c = args[0]
            if c[0] == '#':
                c = c[1:]
            if len(c) == 1:
                (r, g, b, a) = Colors[c]
            if len(c) == 3:
                r = int(c[0]*2, 16)
                g = int(c[1]*2, 16)
                b = int(c[2]*2, 16)
                a = 255
            elif len(c) == 4:
                r = int(c[0]*2, 16)
                g = int(c[1]*2, 16)
                b = int(c[2]*2, 16)
                a = int(c[3]*2, 16)
            elif len(c) == 6:
                r = int(c[0:2], 16)
                g = int(c[2:4], 16)
                b = int(c[4:6], 16)
                a = 255
            elif len(c) == 8:
                r = int(c[0:2], 16)
                g = int(c[2:4], 16)
                b = int(c[4:6], 16)
                a = int(c[6:8], 16)
        elif hasattr(args[0], '__len__'):
            if len(args[0]) == 3:
                (r, g, b) = args[0]
                a = 255
            elif len(args[0]) == 4:
                (r, g, b, a) = args[0]
            elif len(args[0]) == 2:
                return intColor(*args[0])
            else:
                raise Exception(err)
        elif type(args[0]) == int:
            return intColor(args[0])
        else:
            raise Exception(err)
    elif len(args) == 3:
        (r, g, b) = args
        a = 255
    elif len(args) == 4:
        (r, g, b, a) = args
    else:
        raise Exception(err)

    args = [r,g,b,a]
    #args = [0 if np.isnan(a) or np.isinf(a) else a for a in args]
    args = list(map(int, args))
    return QtGui.QColor(*args)

def mkBrush(*args, **kwds):
    """
    | Convenience function for constructing Brush.
    | This function always constructs a solid brush and accepts the same arguments as :func:`mkColor() <pyqtgraph.mkColor>`
    | Calling mkBrush(None) returns an invisible brush.
    """
    if 'color' in kwds:
        color = kwds['color']
    elif len(args) == 1:
        arg = args[0]
        if arg is None:
            return QtGui.QBrush(QtCore.Qt.NoBrush)
        elif isinstance(arg, QtGui.QBrush):
            return QtGui.QBrush(arg)
        else:
            color = arg
    elif len(args) > 1:
        color = args
    return QtGui.QBrush(mkColor(color))
    
def arrayToQPath(x, y, connect='all'):
    """Convert an array of x,y coordinats to QPainterPath as efficiently as possible.
    The *connect* argument may be 'all', indicating that each point should be
    connected to the next; 'pairs', indicating that each pair of points
    should be connected, or an array of int32 values (0 or 1) indicating
    connections.
    """
    
    #CODE IS COPIED FROM PYQTGRAPH

    ## Create all vertices in path. The method used below creates a binary format so that all
    ## vertices can be read in at once. This binary format may change in future versions of Qt,
    ## so the original (slower) method is left here for emergencies:
        #path.moveTo(x[0], y[0])
        #if connect == 'all':
            #for i in range(1, y.shape[0]):
                #path.lineTo(x[i], y[i])
        #elif connect == 'pairs':
            #for i in range(1, y.shape[0]):
                #if i%2 == 0:
                    #path.lineTo(x[i], y[i])
                #else:
                    #path.moveTo(x[i], y[i])
        #elif isinstance(connect, np.ndarray):
            #for i in range(1, y.shape[0]):
                #if connect[i] == 1:
                    #path.lineTo(x[i], y[i])
                #else:
                    #path.moveTo(x[i], y[i])
        #else:
            #raise Exception('connect argument must be "all", "pairs", or array')

    ## Speed this up using >> operator
    ## Format is:
    ##    numVerts(i4)   0(i4)
    ##    x(f8)   y(f8)   0(i4)    <-- 0 means this vertex does not connect
    ##    x(f8)   y(f8)   1(i4)    <-- 1 means this vertex connects to the previous vertex
    ##    ...
    ##    0(i4)
    ##
    ## All values are big endian--pack using struct.pack('>d') or struct.pack('>i')

    path = QtGui.QPainterPath()

    #profiler = debug.Profiler()
    n = x.shape[0]
    # create empty array, pad with extra space on either end
    arr = np.empty(n+2, dtype=[('x', '>f8'), ('y', '>f8'), ('c', '>i4')])
    # write first two integers
    #profiler('allocate empty')
    byteview = arr.view(dtype=np.ubyte)
    byteview[:12] = 0
    byteview.data[12:20] = struct.pack('>ii', n, 0)
    #profiler('pack header')
    # Fill array with vertex values
    arr[1:-1]['x'] = x
    arr[1:-1]['y'] = y

    # decide which points are connected by lines
    #I replaced eq function by ==
    if connect == 'all':
        arr[1:-1]['c'] = 1
    elif connect == 'pairs':
        arr[1:-1]['c'][::2] = 1
        arr[1:-1]['c'][1::2] = 0
    elif connect == 'finite':
        arr[1:-1]['c'] = np.isfinite(x) & np.isfinite(y)
    elif isinstance(connect, np.ndarray):
        arr[1:-1]['c'] = connect
    else:
        raise Exception('connect argument must be "all", "pairs", "finite", or array')

    #profiler('fill array')
    # write last 0
    lastInd = 20*(n+1)
    byteview.data[lastInd:lastInd+4] = struct.pack('>i', 0)
    #profiler('footer')
    # create datastream object and stream into path

    ## Avoiding this method because QByteArray(str) leaks memory in PySide
    #buf = QtCore.QByteArray(arr.data[12:lastInd+4])  # I think one unnecessary copy happens here

    path.strn = byteview.data[12:lastInd+4] # make sure data doesn't run away
    try:
        buf = QtCore.QByteArray.fromRawData(path.strn)
    except TypeError:
        buf = QtCore.QByteArray(bytes(path.strn))
    #profiler('create buffer')
    ds = QtCore.QDataStream(buf)

    ds >> path
    #profiler('load')

    return path    


import numpy as np
from scipy.ndimage.filters import convolve

#https://github.com/colour-science

def as_float_array(array):
    return array.astype('double')

def masks_CFA_Bayer(shape, pattern='RGGB'):
    """
    Returns the *Bayer* CFA red, green and blue masks for given pattern.

    Parameters
    ----------
    shape : array_like
        Dimensions of the *Bayer* CFA.
    pattern : unicode, optional
        **{'RGGB', 'BGGR', 'GRBG', 'GBRG'}**,
        Arrangement of the colour filters on the pixel array.

    Returns
    -------
    tuple
        *Bayer* CFA red, green and blue masks.

    Examples
    --------
    >>> from pprint import pprint
    >>> shape = (3, 3)
    >>> pprint(masks_CFA_Bayer(shape))
    (array([[ True, False,  True],
           [False, False, False],
           [ True, False,  True]], dtype=bool),
     array([[False,  True, False],
           [ True, False,  True],
           [False,  True, False]], dtype=bool),
     array([[False, False, False],
           [False,  True, False],
           [False, False, False]], dtype=bool))
    >>> pprint(masks_CFA_Bayer(shape, 'BGGR'))
    (array([[False, False, False],
           [False,  True, False],
           [False, False, False]], dtype=bool),
     array([[False,  True, False],
           [ True, False,  True],
           [False,  True, False]], dtype=bool),
     array([[ True, False,  True],
           [False, False, False],
           [ True, False,  True]], dtype=bool))
    """

    pattern = pattern.upper()

    channels = dict((channel, np.zeros(shape)) for channel in 'RGB')
    for channel, (y, x) in zip(pattern, [(0, 0), (0, 1), (1, 0), (1, 1)]):
        channels[channel][y::2, x::2] = 1

    return tuple(channels[c].astype(bool) for c in 'RGB')
    
def bayer_split(array, pattern='RGGB'):
    
    assert array.ndim == 2
    
    original_type = array.dtype
    R_m, G_m, B_m = masks_CFA_Bayer(array.shape, pattern)

    R = array * R_m
    G = array * G_m
    B = array * B_m
    
    procarr = np.ndarray(list(array.shape) + [3], original_type)
    procarr[:,:,0] = R
    procarr[:,:,1] = G
    procarr[:,:,2] = B

    return procarr    


def demosaicing_CFA_Bayer_bilinear(CFA, pattern='RGGB'):
    """
    Returns the demosaiced *RGB* colourspace array from given *Bayer* CFA using
    bilinear interpolation.

    Parameters
    ----------
    CFA : array_like
        *Bayer* CFA.
    pattern : unicode, optional
        **{'RGGB', 'BGGR', 'GRBG', 'GBRG'}**,
        Arrangement of the colour filters on the pixel array.

    Returns
    -------
    ndarray
        *RGB* colourspace array.

    Notes
    -----
    -   The definition output is not clipped in range [0, 1] : this allows for
        direct HDRI / radiance image generation on *Bayer* CFA data and post
        demosaicing of the high dynamic range data as showcased in this
        `Jupyter Notebook <https://github.com/colour-science/colour-hdri/\
blob/develop/colour_hdri/examples/\
examples_merge_from_raw_files_with_post_demosaicing.ipynb>`__.

    References
    ----------
    :cite:`Losson2010c`

    Examples
    --------
    >>> import numpy as np
    >>> CFA = np.array(
    ...     [[0.30980393, 0.36078432, 0.30588236, 0.3764706],
    ...      [0.35686275, 0.39607844, 0.36078432, 0.40000001]])
    >>> demosaicing_CFA_Bayer_bilinear(CFA)
    array([[[ 0.69705884,  0.17941177,  0.09901961],
            [ 0.46176472,  0.4509804 ,  0.19803922],
            [ 0.45882354,  0.27450981,  0.19901961],
            [ 0.22941177,  0.5647059 ,  0.30000001]],
    <BLANKLINE>
           [[ 0.23235295,  0.53529412,  0.29705883],
            [ 0.15392157,  0.26960785,  0.59411766],
            [ 0.15294118,  0.4509804 ,  0.59705884],
            [ 0.07647059,  0.18431373,  0.90000002]]])
    >>> CFA = np.array(
    ...     [[0.3764706, 0.360784320, 0.40784314, 0.3764706],
    ...      [0.35686275, 0.30980393, 0.36078432, 0.29803923]])
    >>> demosaicing_CFA_Bayer_bilinear(CFA, 'BGGR')
    array([[[ 0.07745098,  0.17941177,  0.84705885],
            [ 0.15490197,  0.4509804 ,  0.5882353 ],
            [ 0.15196079,  0.27450981,  0.61176471],
            [ 0.22352942,  0.5647059 ,  0.30588235]],
    <BLANKLINE>
           [[ 0.23235295,  0.53529412,  0.28235295],
            [ 0.4647059 ,  0.26960785,  0.19607843],
            [ 0.45588237,  0.4509804 ,  0.20392157],
            [ 0.67058827,  0.18431373,  0.10196078]]])
    """

    original_type = CFA.dtype
    CFA = as_float_array(CFA)
    R_m, G_m, B_m = masks_CFA_Bayer(CFA.shape, pattern)

    H_G = np.array(
        [[0, 1, 0],
         [1, 4, 1],
         [0, 1, 0]], 'double') / 4  # yapf: disable

    H_RB = np.array(
        [[1, 2, 1],
         [2, 4, 2],
         [1, 2, 1]], 'double') / 4  # yapf: disable

    R = convolve(CFA * R_m, H_RB)
    G = convolve(CFA * G_m, H_G)
    B = convolve(CFA * B_m, H_RB)
    
    procarr = np.ndarray(list(CFA.shape) + [3], original_type)
    procarr[:,:,0] = R
    procarr[:,:,1] = G
    procarr[:,:,2] = B

    return procarr


import ctypes
import ctypes.wintypes
from pathlib import Path

from qtpy import QtCore, QtGui, QtWidgets
from qtpy import PYSIDE, PYSIDE2, PYQT4, PYQT5

def using_pyside():
    return PYSIDE or PYSIDE2
    
def using_pyqt():
    return PYQT4 or PYQT5

LASTMAP = None
    
class LASTINPUTINFO(ctypes.Structure):
    
    _fields_ = [
      ('cbSize', ctypes.wintypes.UINT),
      ('dwTime', ctypes.wintypes.DWORD),
      ]

PLASTINPUTINFO = ctypes.POINTER(LASTINPUTINFO)

GetLastInputInfo = ctypes.windll.user32.GetLastInputInfo
GetLastInputInfo.restype = ctypes.wintypes.BOOL
GetLastInputInfo.argtypes = [PLASTINPUTINFO]

def get_last_input_moment():
    liinfo = LASTINPUTINFO()
    liinfo.cbSize = ctypes.sizeof(liinfo)
    GetLastInputInfo(ctypes.byref(liinfo))
    return liinfo.dwTime

class ExecTimeout:
    def __init__(self, dialog, timeout=None):  
        self.dialog = dialog
        self.timer = QtCore.QTimer()
        self.timer.timeout.connect(self.check_user_input_and_close)
        self.timeout = timeout
        
        if not timeout is None:            
            self.timer.start(timeout)
        
    def exec_(self):
        self.exec_user_input_moment = get_last_input_moment()
        self.dialog.exec_()
        self.timer.stop()
        
    def check_user_input_and_close(self):
        last_user_input_moment = get_last_input_moment() 
        
        if self.exec_user_input_moment < last_user_input_moment:
            self.timer.stop()
            
        else:
            self.dialog.close()
           
        
def getFile(filter='*.*', title='open', defaultFile=None, hideFilterDetails=False):
    global LASTMAP

    if not defaultFile is None:
        defaultDir = defaultFile
    else:
        defaultDir = LASTMAP

    #The HideNameFilterDetails option seems to be crashy
    #if the filter doesn't have the detailed form
    #  Tagged Image Format (*.tif)
    #For example the filter *.* often crashes if the option is enabled
    #Maybe QT returns a buggy string as selectedFilter?

    if using_pyside():
        if not hideFilterDetails:
            fileName, selectedFilter = QtWidgets.QFileDialog.getOpenFileName(
                caption = title,
                dir=defaultDir,
                filter=filter)
        else:
            fileName, selectedFilter = QtWidgets.QFileDialog.getOpenFileName(
                caption = title,
                dir=defaultDir,
                filter=filter,
                options=QtWidgets.QFileDialog.HideNameFilterDetails)
                
    elif using_pyqt():
        if not hideFilterDetails:
            fileName, selectedFilter = QtWidgets.QFileDialog.getOpenFileNameAndFilter(
                caption = title,
                directory=defaultDir,
                filter=filter)
        else:
            fileName, selectedFilter = QtWidgets.QFileDialog.getOpenFileNameAndFilter(
                caption = title,
                directory=defaultDir,
                filter=filter,
                options=QtWidgets.QFileDialog.HideNameFilterDetails)

    LASTMAP = str(Path(fileName).parent)
    return fileName, selectedFilter

def getFiles(filter='*.*', title='open', defaultFile=None, hideFilterDetails=False):
    global LASTMAP

    if not defaultFile is None:
        defaultDir = defaultFile
    else:
        defaultDir = LASTMAP

    #The HideNameFilterDetails option seems to be crashy
    #if the filter doesn't have the detailed form
    #  Tagged Image Format (*.tif)
    #For example the filter *.* often crashes if the option is enabled
    #Maybe QT returns a buggy string as selectedFilter?

    if using_pyside():
        if not hideFilterDetails:
            fileNames, selectedFilter = QtWidgets.QFileDialog.getOpenFileNames(
                caption = title,
                dir=defaultDir,
                filter=filter)
        else:
            fileNames, selectedFilter = QtWidgets.QFileDialog.getOpenFileNames(
                caption = title,
                dir=defaultDir,
                filter=filter,
                options=QtWidgets.QFileDialog.HideNameFilterDetails)
    elif using_pyqt():
        if not hideFilterDetails:
            fileNames, selectedFilter = QtWidgets.QFileDialog.getOpenFileNamesAndFilter(
                caption = title,
                directory=defaultDir,
                filter=filter)
        else:
            fileNames, selectedFilter = QtWidgets.QFileDialog.getOpenFileNamesAndFilter(
                caption = title,
                directory=defaultDir,
                filter=filter,
                options=QtWidgets.QFileDialog.HideNameFilterDetails)

    LASTMAP = str(Path(fileNames[0]).parent)
    return fileNames, selectedFilter

def putFile(filter='*.*', title='save', defaultFile=None, defaultFilter=''):
    global LASTMAP

    if not defaultFile is None:
        defaultDir = defaultFile
    else:
        defaultDir = LASTMAP

    if using_pyside():
        fileName, selectedFilter = QtWidgets.QFileDialog.getSaveFileName(
            caption = title,
            filter = filter,
            dir = defaultDir,
            selectedFilter = defaultFilter)
    elif using_pyqt():
        fileName, selectedFilter = QtWidgets.QFileDialog.getSaveFileNameAndFilter(
            caption = title,
            filter = filter,
            directory = defaultDir,
            selectedFilter = defaultFilter)
    LASTMAP = str(Path(fileName).parent)
    return fileName, selectedFilter

def getMap(startPath=None, title='select a Directory'):
    global LASTMAP
    startPath = startPath or LASTMAP
    if using_pyside():
        path = QtWidgets.QFileDialog.getExistingDirectory(
            caption=title,
            dir=startPath)
    elif using_pyqt():
            path = QtWidgets.QFileDialog.getExistingDirectory(
            caption=title,
            directory=startPath)
    LASTMAP = path
    return path

def getString(prompt, default='', title='Input', echo='Normal'):
    """
    Show a popup-window to ask the user some textual input.

    Makes use of QtWidgets.QInputDialog.getText; see
    https://srinikom.github.io/pyside-docs/PySide/QtGui/QInputDialog.html#PySide.QtGui.PySide.QtGui.QInputDialog.getText

    :param str prompt: The explanation that is visible just above the text input field.
    :param str default: The text that is already present in the editable input field.
    :param str title: The name of the pop-window (shown in its title bar).
    :param str echo: 'Normal' for normal text entry; 'Password' for password entry.  See
     http://doc.qt.io/qt-4.8/qlineedit.html#EchoMode-enum
    """
    
    echo_mode = getattr(QtWidgets.QLineEdit.EchoMode, echo)
    return QtWidgets.QInputDialog.getText(None, title, prompt, echo=echo_mode, text=default)[0]
        
               
def getStringTimeout(prompt, default='', title='Input', echo='Normal', timeout=10000):       
    
    echo_mode = getattr(QtWidgets.QLineEdit.EchoMode, echo)
    dialog = QtWidgets.QInputDialog()
    dialog.setWindowTitle(title)
    dialog.setLabelText(prompt)
    dialog.setTextValue(default)    
    
    retval = ExecTimeout(dialog, timeout).exec_()    
    
    if retval == 0:
        return default
        
    return dialog.textValue()
    
def selectFiles(filter='*.*', title='Select', defaultPath=None):
    filedialog = QtWidgets.QFileDialog()
    filedialog.setFileMode(QtWidgets.QFileDialog.FileMode.AnyFile)
    filedialog.setNameFilter(filter)
    
    if not defaultPath is None:
        defaultPath = Path(defaultPath)
        
        if defaultPath.is_file():
            defaultDir = defaultPath.parent
            defaultFile = defaultPath.name
            
        elif defaultPath.is_dir():
            defaultDir = defaultPath
            defaultFile = None
    else:
        defaultDir = None
        defaultFile = None                
    
    if not defaultDir is None:
        filedialog.setDirectory(str(defaultDir))
        
    if not defaultFile is None:
        filedialog.selectFile(str(defaultFile))
        
    filedialog.setWindowTitle(title)
    
    #filedialog.setAcceptMode(QtWidgets.QFileDialog.AcceptOpen)
    
    for btn in filedialog.findChildren(QtWidgets.QPushButton):
        if btn.text() == "&Open":
            btn.setText("Select")
        
    filedialog.exec_()
    return filedialog.selectedFiles()

def getMultiString(prompt, default, title='Input', timeout=None):
    dlg = MultiString(prompt, default, title)
    ExecTimeout(dlg, timeout).exec_()
    return dlg.lns

def messageBox(message, title='', icon='none'):
    icon = icon or 'none'        
    icon = icon.lower()
    if icon == 'help':
        icon = QtWidgets.QMessageBox.Question
    elif icon == 'info':
        icon = QtWidgets.QMessageBox.Information
    elif icon == 'warn':
        icon = QtWidgets.QMessageBox.Warning
    elif icon == 'error':
        icon = QtWidgets.QMessageBox.Critical	
    else:
        icon = QtWidgets.QMessageBox.NoIcon
        
    msgBox = QtWidgets.QMessageBox(icon, title, message)            
    #msgBox.setText(message)
    #msgBox.setIcon(icon)

    msgBox.exec_()
    
def questionBox(question, title=''):
    flags = QtWidgets.QMessageBox.question(None,  title, question, QtWidgets.QMessageBox.Yes, QtWidgets.QMessageBox.No)
    if flags == 16384:
        return True
    else:
        return False        

class MultiString(QtWidgets.QDialog):

    def __init__(self, prompt = [], default = [], title="Input"):
        super().__init__(None)
        self.initui(prompt, default, title)

    def initui(self, prompt, default, title="Input"):
        self.setWindowTitle(title)
        layout = QtWidgets.QVBoxLayout()
        self.edits = []

        for [p,d] in zip(prompt, default):
            pw = QtWidgets.QLabel(p, self)
            dw = QtWidgets.QLineEdit(d, self)
            self.edits.append(dw)
            layout.addWidget(pw)
            layout.addWidget(dw)

        self.okbtn = QtWidgets.QPushButton('Ok')
        self.okbtn.clicked.connect(self.finish)

        layout.addWidget(self.okbtn)

        self.setLayout(layout)

    def finish(self):
        self.lns = []
        for e in self.edits:
            self.lns.append(e.text())
        self.close()


class TopMessageBox(QtWidgets.QWidget):

    def __init__(self):
        super().__init__()
        self.setWindowFlags(self.windowFlags() | QtCore.Qt.WindowStaysOnTopHint)



from qtpy import QtCore, QtGui, QtWidgets

QtSignal = QtCore.Signal

from .point import Point

#Point = QtCore.QPointF

class SceneView(QtWidgets.QGraphicsView):

    scaled = QtSignal(bool, bool)
    scale_ended = QtSignal(bool, bool)
    translated = QtSignal(bool, bool)
    translate_ended = QtSignal(bool, bool)
    zoom_full = QtSignal()
    
    def __init__(self, parent):
        super().__init__(parent)
        #self.setMouseTracking(True)
        #self.setInteractive(False)  
        
        #self.useOpenGL(True)
        
        self.setCacheMode(self.CacheBackground)
        
        self.scale = [1, 1]
        self.center = [0, 0]
        
        self.fixScaleX = False
        self.fixScaleY = False
        
        self.lastMousePos = None        
        
        self.setVerticalScrollBarPolicy(QtCore.Qt.ScrollBarAlwaysOff)
        self.setHorizontalScrollBarPolicy(QtCore.Qt.ScrollBarAlwaysOff)

        self.setFocusPolicy(QtCore.Qt.StrongFocus)        
        self.setFrameShape(QtWidgets.QFrame.NoFrame)
        self.setTransformationAnchor(QtWidgets.QGraphicsView.NoAnchor)
        self.setResizeAnchor(QtWidgets.QGraphicsView.NoAnchor)
        self.setAlignment(QtCore.Qt.AlignLeft | QtCore.Qt.AlignTop)        
        
        self.freeze_y0 = False
        self.old_view_rect = None
        
        # self.restore_view_rect_timer = QtCore.QTimer()
        # self.restore_view_rect_timer.timeout.connect(self.restore_view_rect)
        # self.restore_view_rect_timer.setSingleShot(True)        
        
        self.updateMatrix()

    def useOpenGL(self, b=True):
        if b:
            from PySide2 import QtOpenGL
            v = QtOpenGL.QGLWidget()
        else:
            v = QtWidgets.QWidget()
            
        self.setViewport(v)       

    # def restore_view_rect(self):
        # x0, y0, x1, y1 = self.old_view_rect
        # self.setXLimits(x0, x1, 0,0)
        # self.setYLimits(y0, y1, 0,0)
        # self.old_view_rect = None
        # self.scaled.emit(True, True)
        
    def setRange(self, scene_width, scene_height):
        self.scale[0] = self.width() / scene_width
        self.scale[1] = self.height() / scene_height
        self.updateMatrix()
        
    def updateMatrix(self, propagate=True):        
        #t = self.transform()
        self.limitRefreshRange()        
        self.setTransform(QtGui.QTransform(\
            self.scale[0], 0  , 0,\
            0  , self.scale[1], 0,\
            0  , 0  , 1))
                
        self.centerOn(*self.center)       
        self.viewport().update()

    def translate(self, dx, dy):
        self.center = [self.center[0] +dx, self.center[1] + dy]
        self.updateMatrix()     
        
    def limitRefreshRange(self):
        self.range = QtCore.QRectF()
        self.range.setWidth((self.width() + 20)/ self.scale[0])
        self.range.setHeight((self.height() + 20) / self.scale[1])
        self.range.moveCenter(QtCore.QPointF(*self.center))        
        self.setSceneRect(self.range)          

    def setXPosScale(self, pos, scale):
        self.scale[0] = scale
        self.center[0] = pos + self.width() / 2.0 / scale 
        
        self.limitRefreshRange()
        
        t = self.transform()
        self.setTransform(QtGui.QTransform(\
            scale, 0      , 0,\
            0    , t.m22(), 0,\
            0    , 0      , 1))                    
        self.centerOn(*self.center)       
        
    def setXLimits(self, low, high, left_border=0, right_border=0):
        self.scale[0] = (self.width() - left_border - right_border)/ (high - low)
        self.center[0] = (low - left_border / self.scale[0] + high) / 2 
        
        self.limitRefreshRange()
        self.updateMatrix()              

    def setYPosScale(self, pos, scale):
        self.scale[1] = scale
        self.center[1] = pos + self.height() / 2.0 / scale 
        
        self.limitRefreshRange()
        
        t = self.transform()
        self.setTransform(QtGui.QTransform(\
            t.m11(), 0    , 0,\
            0      , scale, 0,\
            0      , 0    , 1))                    
        self.centerOn(*self.center)  
        
    def setYLimits(self, low, high, bottom_border=0, top_border=0):
        self.scale[1] = (self.height() - bottom_border - top_border) / (low - high)
        self.center[1] = (low + bottom_border / self.scale[1] + high) / 2 
        
        self.limitRefreshRange()
        self.updateMatrix()            

    def wheelEvent(self, ev):
        #QtWidgets.QGraphicsView.wheelEvent(self, ev)
        if not self.fixScaleX and (self.fixScaleY or self.freeze_y0):                
            self.scale[0] = self.scale[0] * 1.001 ** ev.delta()        
            self.updateMatrix()  
            self.scale_ended.emit(True, False)
                
        elif self.fixScaleX and not self.fixScaleY:                
            self.scale[1] = self.scale[1] * 1.001 ** ev.delta()
            self.updateMatrix()
            self.scale_ended.emit(False, True)                
                
        else:
            self.scale[0] = self.scale[0] * 1.001 ** ev.delta()
            self.scale[1] = self.scale[1] * 1.001 ** ev.delta()
            self.updateMatrix()
            self.scale_ended.emit(True, True)
    
    def pixelSize(self):
        """Return vector with the length and width of one view pixel in scene coordinates"""
        p0 = Point(0,0)
        p1 = Point(1,1)
        tr = self.transform().inverted()[0]
        p01 = tr.map(p0)
        p11 = tr.map(p1)
        return Point(p11 - p01) 

    def mousePressEvent(self, ev):    
        self.lastMousePos = Point(ev.pos())
        self.lastPressButton = ev.button()
        super().mousePressEvent(ev)
        #print('mouse press x:%d y:%d' % (self.lastMousePos[0], self.lastMousePos[1]))
        
    def mouseDoubleClickEvent(self, ev):            
        self.zoom_full.emit()
               
    def mouseMoveEvent(self, ev):               
        super().mouseMoveEvent(ev)
        
        if hasattr(self.scene(), 'moving_indicators') and self.scene().moving_indicators:
            self.scene().moving_indicators = False
            return
        
        if self.lastMousePos is None:
            self.lastMousePos = Point(ev.pos())
            
        delta = Point(ev.pos()) - self.lastMousePos
        self.lastMousePos = Point(ev.pos())
        
        if ev.buttons() in [QtCore.Qt.LeftButton]:
        #if ev.buttons() in [QtCore.Qt.MidButton, QtCore.Qt.RightButton]: 
            px = self.pixelSize()
            tr = -delta * px
            movex, movey = True, True
            if self.fixScaleX:
                tr[0] = 0
                movex = False
            if self.fixScaleY or self.freeze_y0:
                tr[1] = 0
                movey = False
            self.translate(tr[0], tr[1])   
            self.translated.emit(movex, movey)
        elif ev.buttons() in [QtCore.Qt.RightButton]:
            if not self.freeze_y0:
                px = self.pixelSize()
                self.scale[0] = self.scale[0] * 1.01 ** delta[0]
                self.scale[1] = self.scale[1] * 1.01 ** delta[1]
                self.updateMatrix()
                self.scaled.emit(True, True)
            elif self.freeze_y0:
                scale_x = 1.01 ** (-delta[0])
                scale_y = 1.01 ** delta[1]
                x0, y0, x1, y1 = self.viewRectCoord()
                
                cx = (x1 + x0) / 2
                rx = (x1 - x0) / 2
                x0 = cx - rx * scale_x
                x1 = cx + rx * scale_x
                self.setXLimits(x0, x1, 0, 0)
                
                y0 = y0 * scale_y
                y1 = y1 * scale_y                
                self.setYLimits(y0, y1, 0, 0)                
                
                self.scaled.emit(True, True)
        
    def viewRectCoord(self):
        #return the currently visible window in view coord
        xc, yc = self.center
        x0 = xc - self.width() / abs(self.scale[0]) / 2
        y0 = yc - self.height() / abs(self.scale[1]) / 2
        x1 = xc + self.width() / abs(self.scale[0]) / 2
        y1 = yc + self.height() / abs(self.scale[1]) / 2
        return x0, y0, x1, y1

    def refresh(self):
        self.viewport().update()
        
    def resizeEvent(self, ev):
        # if self.old_view_rect is None:
            # self.old_view_rect = x0, y0, x1, y1 = self.viewRectCoord()
            # self.restore_view_rect_timer.start(500)        
        self.updateMatrix()


import sys, io, os, builtins
import json
import signal
import traceback
import threading
import multiprocessing
import textwrap
import types
import pathlib
import time
import pickle
import configparser
import time
import collections
import platform
import logging
from multiprocessing import Process, Lock
from queue import Queue
    
try:
    import zmq
    zmq_context = zmq.Context()
    
except:
    pass    


logger = logging.getLogger(__name__)   

#In case of Process call, this module will be important by the child 
#process by the unpickling of the CommQueues
from .conf import config, configure

#Configure in case of the child process before importing anthing else of ghawk2
configure(matplotlib={'backend':'svg'})
process_name = multiprocessing.current_process().name
logger.debug(f'import of {__name__} by {process_name}\n')

from .shellmod import Shell
from .gui_proxy import GuiProxy
from .interpreter import Interpreter, QueueInterpreter
from .comm import NonDuplexQueue, ZmqQueues, CommQueues

PROCESSES = collections.OrderedDict()

here = pathlib.Path(__file__).parent  

        
class TimeOutGuiCall(object):
    def __init__(self, gui_proxy, timeout, callback, args=()):
        self.gui_proxy = gui_proxy
        self.callback = callback
        self.called = False
        self.timeout = timeout
        self.thread = threading.Thread(target=self.delay_call, args=args)
        self.thread.start()
        self.lock = threading.Lock()
        
    def delay_call(self, *args):
        time.sleep(self.timeout)
        if self.called: return
        try:
            self.lock.acquire()
            self.called = True
            self.gui_proxy._call(self.callback, *args)
        finally:
            self.lock.release()
        
    def call(self, *args):
        if self.called: return
        try:
            self.lock.acquire()
            self.called = True
            self.callback(*args)
        finally:
            self.lock.release()
        

class TaskBase(object):   
    def __init__(self, tasktype='process'):
        self.tasktype = tasktype
        self.process_id = -1
        self.thread_name = 'invalid'
        self.thread_id = -1
        
    @property
    def stdin_queue(self):
        return self.cqs.stdin_queue        
    
    @property
    def stdout_queue(self):
        return self.cqs.stdout_queue
        
    @property
    def cmd_queue(self):
        return self.cqs.cmd_queue     

    @property
    def return_queue(self):
        return self.cqs.return_queue      

    def is_current_thread(self):
        return self.process_id == os.getpid() and self.thread_id == threading.get_ident()          

    def getReturnedValues(self, mode, error_code, result):
        if error_code == 0:
            pass
        else:
            print(f'WARNING with error code {error_code}\n{mode} {result}')
            
        return mode, error_code, result

    def send_command(self, command, callback=None):
        return self.send_func_and_call('interprete', (command,), callback)
        
    def send_input(self, text):
        return self.send_func_and_call('input', (text,))
        
    def evaluate(self, *args, multiple=False):
        if multiple:
            return self.send_func_and_call('eval', args, wait=True)
        else:
            result = self.send_func_and_call('eval', args, wait=True)
            return result[0]
        
    def call_func(self, func, args=(), callback=None, wait=False, queue='stdin'):           
        if not isinstance(func, str):
            func = self.gui_proxy.encode_func(func)
        
        if queue == 'stdin':
            return self.send_func_and_call('func', (func, args), callback, wait)
        elif queue == 'flow':
            return self.send_func_and_call('flow_func', (func, args), callback, wait)
            
    def send_func_and_call(self, mode, args=(), callback=None, wait=False, timeout=0):
        if callback is None and not wait:
            callback = self.getReturnedValues    
    
        if not callback is None:
            if timeout == 0:
                call_back_id = self.gui_proxy.encode_func(callback, register=True)  
                
            else:
                callbackwrap = TimeOutGuiCall(self.gui_proxy, timeout, callback, args=('timeout', 5, None))                
                call_back_id = self.gui_proxy.encode_func(callbackwrap.call, register=True)
        else:
            call_back_id = None
            
        if mode in ['input', 'interprete', 'func', 'console id', 'exit']:
            self.stdin_queue.put((mode, args, call_back_id))
            
        elif mode in ['flow', 'eval', 'flow_func']:        
            self.cmd_queue.put((mode, args, call_back_id))
        
        if self.is_current_thread():
            self.mainshell.interpreters[self.thread_id].execute()

        if callback is None and wait:
            #This is an high risk for a deathlock
            #Waiting here will block the eventloop
            #If still prior callbacks are queued, the eventloop can not respond to it -> deathlock     
            
            mode, error_code, result = self.return_queue.get(timeout=5)               
                
            assert error_code == 0
            return result               
        
    def register(self, mainshell):
        #mainshell.tasks[(self.process_id, self.thread_id)] = self
        pass
        
    def print_trace(self):
        self.send_func_and_call("flow", ("trace",))        
        
    def print_locals(self):
        self.send_func_and_call("flow", ("locals",))                
        
    def sync_break(self):
        self.send_func_and_call("flow", ("sync_break",))
        
    def async_break(self):
        self.send_func_and_call("flow", ("KeyboardInterrupt",))  
        
    def system_exit(self):
        self.send_func_and_call("flow", ("system_exit",))          
        
    def kill(self):
        import psutil
        
        parent = psutil.Process(self.process_id)
        for child in parent.children(recursive=True):
            child.kill()
            
        parent.kill()
        
        tasks = list(PROCESSES[self.process_id].values())
        for task in tasks:
            task.console.close_panel()            
        
    def flow_alive(self, callback, timeout=5):        
        self.send_func_and_call("flow", ("heartbeat",), callback, timeout=timeout)        
        
    def set_tracing(self, enable=True):
        self.send_func_and_call("flow", ("set_tracing", enable))
        
    def set_timeit(self, enable=True):
        self.send_func_and_call("flow", ("set_timeit", enable))
        
    def enable_profiling(self):
        self.send_func_and_call("flow", ("enable_profiling",))        
        
    def process_ready(self, *args):
        #Gui will freeze until something comes back from the return_queue
        anstype, error_code, (pid, tname, tid) = args
        
        self.process_id = pid
        self.thread_name = tname
        self.thread_id = tid
        
        if not self.process_id in PROCESSES:
            PROCESSES[self.process_id] = dict()
            
        PROCESSES[self.process_id][self.thread_id] = self
        
        #This can maybe come to soon, before the console was created
        if hasattr(self, 'console'):
            self.console.refresh_pid_tid()
            self.send_func_and_call("console id", (self.console.panid,))
            
    def unregister(self):
        PROCESSES[self.process_id].pop(self.thread_id)
        if len(PROCESSES[self.process_id]) == 0:
            PROCESSES.pop(self.process_id)

    def set_flusher(self, func):
        self.gui_proxy.set_func_hook(-1, func)  
        
    def wait_process_ready(self, timeout=3):
        from qtpy import QtWidgets
        
        qapp = QtWidgets.QApplication.instance()
        for i in range(timeout * 10):
            qapp.processEvents()
            if self.thread_name != 'invalid': break
            time.sleep(0.1)


class ThreadTask(TaskBase):  
    cmd_queues = dict()    
    
    def __init__(self, mainshell, new_thread=True):    
        self.new_thread = new_thread    
        
        if new_thread:
            super().__init__('thread')    
        else:
            super().__init__('main')    

        self.mainshell = mainshell        

        self.cqs = CommQueues(Queue)

        self.gui_proxy = GuiProxy(mainshell._qapp,
            self.cqs.gui_call_queue,   #None
            self.cqs.gui_return_queue, #None
            self.process_ready)

    def start(self):
        if self.new_thread:                              
            self.mainshell.new_interactive_thread(self.cqs, self.gui_proxy)

        else:
            self.gui_proxy.block = False
            
            self.thread = threading.currentThread()   
            self.command_loop(self.cqs, self.gui_proxy)        

    def finish(self, close=False):
        if self.thread_name == 'MainThread':
            print("You can't finish the meain thread")
            return
            
        super().finish(close)        
        
    def command_loop(self, cqs, gui_proxy=None):        

        thread_id = threading.get_ident()
        ThreadTask.cmd_queues[thread_id] = cqs.cmd_queue
        
        QueueInterpreter(self.mainshell, cqs, gui_proxy)


class ProcessTask(TaskBase):
    def __init__(self, mainshell, cqs=None):
        super().__init__('child')
        
        if cqs is None:
            self.cqs = CommQueues(multiprocessing.Queue, process=True)
            self.start_child = True
        else:
            self.cqs = cqs            
            self.start_child = False            
        
        self.gui_proxy = GuiProxy(mainshell._qapp,
            self.cqs.gui_call_queue,
            self.cqs.gui_return_queue,
            self.process_ready)        
        
    def start(self):
        #no existing queue from an existing master process
        #Start a new child process
        if self.start_child:
            self.process = Process(target=ProcessTask.start_child_process, args=(self.cqs, ), daemon=True)
            self.process.start()                   
        
        self.flusher = None
                
    @staticmethod    
    def start_child_process(cqs):            
        try:
            shell = Shell()
            shell.start_in_this_thread(cqs)
            
        finally:
            import psutil
            sys.__stdout__.write(f'End of start_child_processs\n')
            sys.__stdout__.flush()
            proc = psutil.Process()
            proc.kill()

class ProcessThreadTask(TaskBase):
    def __init__(self, mainshell, master_process_task, queue_type='pipe'):
        super().__init__('child-thread')
        self.master_process_task = master_process_task
        
        if queue_type == 'pipe':
            self.cqs = CommQueues(NonDuplexQueue, process=True)          
            
        elif queue_type == 'zmq':
            self.cqs = ZmqQueues()
            self.cqs.setup_as_server()
            
        self.gui_proxy = GuiProxy(mainshell._qapp, self.cqs.gui_call_queue, self.cqs.gui_return_queue)        
        self.gui_proxy.set_func_hook(-2, self.process_ready)
          
    def start(self):          
        #Problems if master_process_task didn't not yet finished prior command
        #For example master_process_task is still queues flushes of a big print loop
        #The master_process_task.return_queue still contains the prior return result
        self.master_process_task.call_func(Shell.new_interactive_thread, args=(self.cqs,), queue='flow')
        



import threading
import sys, os
import ctypes
from collections import OrderedDict
import logging
import importlib
import pprint
from pathlib import Path

import numpy as np

from qtpy import QtGui, QtWidgets, QtCore
from qtpy.QtCore import Qt

from .. import config, gui, __release__, PROGNAME, DOC_HTML
from ..core import conf

from .about import AboutScreen
from ..panels.base import BasePanel
from ..ezdock.laystruct import LayoutStruct
from ..dicttree.widgets import DictionaryTreeDialog

logger = logging.getLogger(__name__)
respath = Path(config['respath'])

class NewPanelMenu(QtWidgets.QMenu):
    def __init__(self, parent=None, showIcon=False):
        super().__init__('New', parent)
        if showIcon:
            self.setIcon(QtGui.QIcon(str(respath / 'icons' / 'px16' / 'application_add.png')))

    @property
    def panels(self):
        return QtWidgets.QApplication.instance().panels

    def showEvent(self, event):
        self.showpos = QtGui.QCursor().pos()
        self.initactions()

    def initactions(self):
        self.clear()
        panelClasses = BasePanel.userPanelClasses()

        self.liveActions = []

        for category, catPanelClasses in panelClasses.items():
            catMenu = QtWidgets.QMenu(category)
            self.addMenu(catMenu)
            for panelClass in catPanelClasses:
                if not panelClass.userVisible: continue
                try:
                    panelShortName =  panelClass.panelShortName
                except:
                    panelShortName = 'unkown'

                action = QtWidgets.QAction(f'{panelShortName} <{panelClass.__qualname__}>')

                if hasattr(panelClass, 'classIconFile'):
                    action.setIcon(QtGui.QIcon(panelClass.classIconFile))

                action.triggered.connect(CachedArgCall(self.newPanel, panelClass, self.parent().windowName, self.showpos))
                catMenu.addAction(action)
                self.liveActions.append(action)

    def newPanel(self, panelClass, windowName, showpos=None):
        if panelClass.panelCategory == 'plot':
            import pylab
            fig = pylab.figure()
        else:
            self.panels.new_panel(panelClass, windowName)


class ShowMenu(QtWidgets.QMenu):
    def __init__(self, parent=None, showIcon=False):
        super().__init__('Panel', parent)
        if showIcon:
            self.setIcon(QtGui.QIcon(str(respath / 'icons' / 'px16' / 'application_get.png')))

    def showEvent(self, event):
        self.initactions()

    @property
    def panels(self):
        return QtWidgets.QApplication.instance().panels

    def preview(self):
        self.previews = PanelsFloatingPreviews()
        self.previews.preview()
        self.previews.exec_()

        self.previews.selectedPanel.select()
        self.previews.selectedPanel.show_me()

    def initactions(self):
        self.clear()

        self.liveActions = []

        action = QtWidgets.QAction(f'Previews...\t{config["shortcuts"]["panel preview"]}', triggered=self.preview)
        self.addAction(action)
        self.liveActions.append(action)
        self.addSeparator()

        for category in self.panels.keys():
            panels = self.panels[category]
            selected_panid = self.panels.selected(category, panel=False)
            for panid in sorted(panels.keys()):

                panel = self.panels[category][panid]
                action = QtWidgets.QAction(panel.windowTitle())
                action.triggered.connect(CachedArgCall(self.showPanel, panel))
                action.setCheckable(True)

                if panel.panid == selected_panid:
                    action.setChecked(True)
                else:
                    action.setChecked(False)

                self.addAction(action)
                self.liveActions.append(action)

    def showPanel(self, panel):
        panel.show_me()
        panel.select()

class WindowMenu(QtWidgets.QMenu):
    def __init__(self, parent=None, showIcon=False):
        super().__init__('Window', parent)
        if showIcon:
            self.setIcon(QtGui.QIcon(str(respath / 'icons' / 'px16' / 'application_double.png')))

    @property
    def windows(self):
        return QtWidgets.QApplication.instance().windows

    def showEvent(self, event):
        self.initactions()

    def preview(self):
        self.previews = WindowsFloatingPreviews()
        self.previews.preview()
        self.previews.exec_()

    def initactions(self):
        self.clear()

        self.liveActions = []

        action = QtWidgets.QAction(f'Previews...\t{config["shortcuts"]["window preview"]}', triggered=self.preview)
        self.addAction(action)
        self.liveActions.append(action)
        self.addSeparator()

        for window_name in self.windows.keys():
            window = self.windows[window_name]
            action = QtWidgets.QAction(window_name)
            action.triggered.connect(CachedArgCall(self.showWindow, window))
            self.addAction(action)
            self.liveActions.append(action)

    def showWindow(self, window):
        window.show()
        window.raise_()


class CachedArgCall(object):
    def __init__(self, caller, *args, **kwargs):
        self.caller = caller
        self.args = args
        self.kwargs = kwargs

    def __call__(self):
        self.caller(*self.args, **self.kwargs)


class LayoutMenu(QtWidgets.QMenu):
    def __init__(self, parent=None, showIcon=False):
        super().__init__('Layout', parent)
        if showIcon:
            self.setIcon(QtGui.QIcon(str(respath / 'icons' / 'px16' / 'layout_content.png')))

        self.initactions()

    @property
    def panels(self):
        return QtWidgets.QApplication.instance().panels

    def showEvent(self, event):
        self.initactions()

    def initactions(self):
        self.clear()

        action = QtWidgets.QAction('Save Layout...', self, triggered=self.saveLayout)
        self.addAction(action)
        self.addSeparator()

        self.addLayoutActions(self)

    def addLayoutActions(self, parent=None):
        shortcuts = dict((v,k) for k,v in config['shortcuts']['layout'].items())
        prefix = config['shortcuts']['layout']['prefix']

        for name, layout in config['layout'].items():
            shortcut = shortcuts.get(name, None)
            if shortcut is None:
                action = QtWidgets.QAction(name, parent)
            else:
                action = QtWidgets.QAction(f'{name}\t{prefix}{shortcut}', parent)
            caller = self.panels.restore_state_from_config
            action.triggered.connect(CachedArgCall(caller, name))
            parent.addAction(action)

    def saveLayout(self):
        layout_name = gui.dialog.getstring('Give it a name')
        if layout_name == '': return
        if layout_name == 'base':
            gui.dialog.msgbox(f"You can't overwrite {layout_name}", icon='warn')
        else:
            config['layout'][layout_name] = gui.qapp.panels.ezm.get_perspective()



class MainDialog(QtWidgets.QMainWindow):
    def __init__(self, panels):
        super().__init__()
        self.setWindowTitle(f'{PROGNAME} {__release__}')
        self.panels = panels
        self.tabs = QtWidgets.QTabWidget(self)
        self.setCentralWidget(self.tabs)

        self.panels_layout = PanelsLayout(self, panels)
        self.tabs.addTab(self.panels_layout, 'Layout')

        self.initMenu()
        self.callerWindow = None

    @property
    def qapp(self):
        return QtWidgets.QApplication.instance()

    @property
    def windowName(self):
        return None

    def initMenu(self):
        self.appMenu = self.menuBar().addMenu("&Application")

        act = QtWidgets.QAction("Restart", self,
            triggered=self.restart,
            statusTip=f"Restart {PROGNAME}")
        self.appMenu.addAction(act)

        act = QtWidgets.QAction("Exit", self, shortcut=QtGui.QKeySequence.Quit,
            statusTip=f"Exit {PROGNAME}",
            triggered=self.qapp.quit,
            icon=QtGui.QIcon(str(respath / 'icons' / 'px16' / 'door_out.png')))

        self.appMenu.addAction(act)

        self.newMenu = NewPanelMenu(self)
        self.menuBar().addMenu(self.newMenu)

        self.showMenu = ShowMenu(self)
        self.menuBar().addMenu(self.showMenu)

        self.windowMenu = WindowMenu(self)
        self.menuBar().addMenu(self.windowMenu)

        self.layoutMenu = LayoutMenu(self)
        self.menuBar().addMenu(self.layoutMenu)

        self.configMenu = self.menuBar().addMenu("Config")
        self.configMenu.addAction(QtWidgets.QAction("View Config", self, triggered=self.showConfig,
            icon=QtGui.QIcon(str(respath / 'icons' / 'px16' / 'page_gear.png'))))

        self.configMenu.addAction(QtWidgets.QAction("Save Config", self, triggered=self.saveConfig))

        self.helpMenu = self.menuBar().addMenu("Help")

        helpAct = QtWidgets.QAction("&Help", self, triggered=self.help)
        helpAct.setIcon(QtGui.QIcon(str(respath / 'icons' / 'px16' / 'help.png')))
        self.helpMenu.addAction(helpAct)

        aboutGhQtAct = QtWidgets.QAction(f"About {PROGNAME}", self, triggered=self.about)
        self.helpMenu.addAction(aboutGhQtAct)

        self.helpMenu.addAction(QtWidgets.QAction("License", self, triggered=self.license))

        infoGhQtAct = QtWidgets.QAction("Instance Info", self, triggered=self.info)
        infoGhQtAct.setIcon(QtGui.QIcon(str(respath / 'icons' / 'px16' / 'information.png')))
        self.helpMenu.addAction(infoGhQtAct)

        aboutQtAct = QtWidgets.QAction("About Qt", self, triggered=self.qapp.aboutQt)
        self.helpMenu.addAction(aboutQtAct)

    def refresh(self):
        self.panels_layout.refresh()

    def exec_(self, callerWindow=None):
        self.callerWindow = callerWindow
        self.raise_()
        self.qapp.setActiveWindow(self)
        self.showNormal()

    def accept(self):
        self.showMinimized()

    def restart(self):
        #self.qapp.quit()
        os.execlp(sys.executable, 'python', *sys.argv)

    def showConfig(self):
        dt = DictionaryTreeDialog(config)
        dt.edit()

    def saveConfig(self):
        path = gui.putfilename('JSON (*.json)', file=config['save_config_file'])
        conf.save_config_json(path)

    def help(self):
        print("Opening %s" % DOC_HTML)
        os.system('start "help" "%s"' % DOC_HTML)

    def about(self):
        aboutScreen = AboutScreen()
        aboutScreen.exec_()

    def license(self):
        message = open(respath / 'LICENSE.txt', 'r').read()
        print(message)
        self.qapp.panels['console'][0].show_me()

    def info(self):
        message = self.qapp.cmdserver.host_info()
        print(message)
        self.qapp.panels['console'][0].show_me()

    def closeEvent(self, event):
        allHidden = True
        for window in self.qapp.windows.values():
            if window.isVisible():
                allHidden = False
                break

        if allHidden:
            event.accept()

        else:
            self.showMinimized()
            self.callerWindow = None
            event.ignore()


class PanelsFloatingPreviews(QtWidgets.QDialog):
    def __init__(self):
        super().__init__()
        self.thumbs = []

        vbox = QtWidgets.QVBoxLayout()
        self.setLayout(vbox)

        font = self.font()
        font.setPointSize(font.pointSize() * 2)

        self.caption = QtWidgets.QLabel('Panels')
        self.caption.setFont(font)
        self.caption.setAlignment(Qt.AlignHCenter | Qt.AlignVCenter)
        vbox.addWidget(self.caption)

        self.boxlay = QtWidgets.QGridLayout()
        vbox.addLayout(self.boxlay)

        self.setWindowFlag(Qt.FramelessWindowHint)
        self.setWindowFlag(Qt.WindowStaysOnTopHint)

    @property
    def panels(self):
        return QtWidgets.QApplication.instance().panels

    def preview(self):
        total = sum((len(pans) for cat, pans in self.panels.items()))
        colcount = int((total*16/9)**0.5)

        index = 0
        for cat in self.panels.keys():
            selectedId = self.panels.selected(cat, panel=False)
            for panid in sorted(self.panels[cat].keys()):
                panel = self.panels[cat][panid]
                pixmap = panel.grab().scaled(160, 160, QtCore.Qt.KeepAspectRatio, QtCore.Qt.SmoothTransformation)

                thumb = QtWidgets.QToolButton()
                thumb.setToolButtonStyle(QtCore.Qt.ToolButtonTextUnderIcon)
                thumb.setIcon(QtGui.QIcon(pixmap))
                thumb.setIconSize(pixmap.rect().size())
                thumb.setText(panel.short_title)
                if panel.panid == selectedId:
                    thumb.setDown(True)
                thumb.setToolTip(panel.long_title)
                thumb.clicked.connect(CachedArgCall(self.showPanel, panel))

                self.thumbs.append(thumb)

                self.boxlay.addWidget(thumb, index // colcount, index % colcount)

                index += 1

    def showPanel(self, panel):
        for thumb in self.thumbs:
            thumb.setParent(None)
            thumb.hide()

        self.thumbs = []
        self.hide()

        self.selectedPanel = panel

class WindowsFloatingPreviews(QtWidgets.QDialog):
    def __init__(self):
        super().__init__()
        self.thumbs = []

        self.setLayout(QtWidgets.QVBoxLayout())

        font = self.font()
        font.setPointSize(font.pointSize() * 2)

        self.caption = QtWidgets.QLabel('Windows')
        self.caption.setFont(font)
        self.caption.setAlignment(Qt.AlignHCenter | Qt.AlignVCenter)
        self.layout().addWidget(self.caption)

        self.boxlay = QtWidgets.QGridLayout()
        self.layout().addLayout(self.boxlay)

        self.setWindowFlag(Qt.FramelessWindowHint)
        self.setWindowFlag(Qt.WindowStaysOnTopHint)

    @property
    def windows(self):
        return QtWidgets.QApplication.instance().windows

    def preview(self):
        total = len(self.windows.items())
        colcount = int((total*16/9)**0.5)

        index = 0
        for window in self.windows.values():
            pixmap = window.grab().scaled(160, 160, QtCore.Qt.KeepAspectRatio, QtCore.Qt.SmoothTransformation)

            thumb = QtWidgets.QToolButton()
            thumb.setToolButtonStyle(QtCore.Qt.ToolButtonTextUnderIcon)
            thumb.setIcon(QtGui.QIcon(pixmap))
            thumb.setIconSize(pixmap.rect().size())
            #thumb.setText(window.windowTitle())
            thumb.setText(window.name)
            thumb.setToolTip(window.windowTitle())
            thumb.clicked.connect(CachedArgCall(self.showWindow, window))

            self.thumbs.append(thumb)

            self.boxlay.addWidget(thumb, index // colcount, index % colcount)

            index += 1

        #self.show()

    def showWindow(self, window):
        window.showNormal()
        window.raise_()

        for thumb in self.thumbs:
            thumb.hide()

        self.thumbs = []
        self.hide()


class LayoutList(QtWidgets.QListWidget):
    def __init__(self, parent):
        super().__init__(parent=parent)
        self.currentItemChanged.connect(self.changeItem)

    def changeItem(self, item):
        if item is None: return
        layout = config['layout'][item.name]
        text = ''
        ls = LayoutStruct()
        for window in layout["windows"]:
            text += f'window: {window["name"]}\n'
            ls.root = window["docks"]
            text += ls.describe() + '\n\n'
        self.parent().parent().preview.setPlainText(text)


class PanelsLayout(QtWidgets.QWidget):
    def __init__(self, dialog, panels):
        super().__init__(parent=dialog)
        self.dialog = dialog
        self.panels = panels
        self.layout_list = LayoutList(self)
        self.preview = QtWidgets.QPlainTextEdit(self)
        console_font = QtGui.QFont('Consolas', pointSize=config['console']['fontsize'])
        self.preview.setFont(console_font)
        self.preview.setWordWrapMode(QtGui.QTextOption.NoWrap)

        self.vbox = QtWidgets.QVBoxLayout()
        self.setLayout(self.vbox)

        self.box = QtWidgets.QSplitter(QtCore.Qt.Horizontal)
        self.vbox.addWidget(self.box)

        self.box.addWidget(self.layout_list)
        self.box.addWidget(self.preview)

        self.loadBtn = QtWidgets.QPushButton('Load')
        self.loadBtn.clicked.connect(self.loadLayout)

        hbox = QtWidgets.QHBoxLayout()
        hbox.addWidget(self.loadBtn)
        self.vbox.addLayout(hbox)
        self.refresh()

    def refresh(self):
        self.layout_list.clear()
        shortcuts = dict((v,k) for k,v in config['shortcuts']['layout'].items())
        for name, layout in config['layout'].items():
            description = layout.get('description', 'no description')
            if shortcuts.get(name, None):
                description += f"\n    [Ctrl+F{shortcuts.get(name, None)}]"
            item = QtWidgets.QListWidgetItem(f'{name}:\n     {description}')
            item.name = name
            self.layout_list.addItem(item)

    def loadLayout(self):
        item = self.layout_list.selectedItems()[0]
        self.panels.restore_state_from_config(item.name)
        self.dialog.accept()




import unittest

class GammaDeskSuite(unittest.TestCase):
    panid = 1

    def test_screenstate_1(self):
        from gdesk import gui
        from pylab import plt
        from pathlib import Path

        gui.load_layout('console')

        samplePath = Path(r'./samples')

        gui.img.select(1)
        gui.img.open(samplePath / 'kodim05.png')
        gui.img.zoom_fit()
        plt.plot(gui.vs.mean(2).mean(1))
        plt.title('Column means of image 1')
        plt.xlabel('Column Number')
        plt.ylabel('Mean')
        plt.grid()
        plt.show()

        gui.img.select(2)
        gui.img.open(samplePath / 'kodim23.png')
        gui.img.zoom_full()
        plt.figure()
        plt.plot(gui.vs.mean(2).mean(0))
        plt.title('Row means of image 2')
        plt.xlabel('Row Number')
        plt.ylabel('Mean')
        plt.grid()
        plt.show()


    def test_small_loop_and_print(self):
        import time
        import sys
        from gdesk import gui

        gui.clc()

        expectedOutput = ''

        for i in range(42):
            line = f'i = {i}'
            print(line)
            expectedOutput += f'{line}\n'
            time.sleep(0.01)

        sys.stdout.flush()
        text = gui.console.text()
        assert expectedOutput == text


    def test_menu_file(self):
        from gdesk import gui

        gui.load_layout('image, levels & console')
        gui.img.select(1)

        gui.img.menu(['File', 'New...'], 1920, 1080, 4, 'uint8', 127)

        self.assertEqual(gui.vs.shape[0], 1080)
        self.assertEqual(gui.vs.shape[1], 1920)
        self.assertEqual(gui.vs.shape[2], 4)
        self.assertEqual(gui.vs.dtype, 'uint8')
        self.assertEqual(gui.vs.min(), 127)
        self.assertEqual(gui.vs.max(), 127)

        gui.img.menu(['File', 'New...'], 3840, 2160, 1, 'uint16', 2**16-1)

        self.assertEqual(gui.vs.shape[0], 2160)
        self.assertEqual(gui.vs.shape[1], 3840)
        self.assertEqual(len(gui.vs.shape), 2)
        self.assertEqual(gui.vs.dtype, 'uint16')
        self.assertEqual(gui.vs.min(), 2**16-1)
        self.assertEqual(gui.vs.max(), 2**16-1)

        gui.img.menu(['File', 'New...'], 640, 480, 3, 'double', 0.5)

        self.assertEqual(gui.vs.shape[0], 480)
        self.assertEqual(gui.vs.shape[1], 640)
        self.assertEqual(gui.vs.shape[2], 3)
        self.assertEqual(gui.vs.dtype, 'double')
        self.assertEqual(gui.vs.min(), 0.5)
        self.assertEqual(gui.vs.max(), 0.5)

        #https://imageio.readthedocs.io/en/stable/standardimages.html
        gui.img.menu(['File', 'Open Image...'], 'imageio:astronaut.png')
        gui.img.menu(['File', 'Open Image...'], 'imageio:wood.jpg')
        gui.img.menu(['File', 'Open Image...'], 'imageio:camera.png')

        gui.img.cmap('turbo')


    def test_menu_canvas(self):
        from gdesk import gui
        import scipy.misc

        arr = scipy.misc.face()
        height, width = arr.shape[:2]

        gui.load_layout('image, levels & console')
        gui.img.select(1)
        gui.show(arr)
        gui.img.zoom_fit()
        gui.menu_trigger('image', None, ['Canvas', 'Flip Horizontal'])
        gui.menu_trigger('image', None, ['Canvas', 'Flip Vertical'])
        gui.menu_trigger('image', None, ['Canvas', 'Rotate Left 90'])
        gui.menu_trigger('image', None, ['Canvas', 'Rotate Right 90'])
        gui.menu_trigger('image', None, ['Canvas', 'Rotate 180'])

        gui.menu_trigger('image', None, ['Canvas', 'Resize Canvas...'], width+64, height+32)
        gui.menu_trigger('image', None, ['Canvas', 'Resize Canvas...'], width, height)

        assert (arr == gui.vs).all()


    def test_menu_view(self):
        gui.load_layout('image, levels & console')
        gui.img.select(1)

        gui.img.menu(['File', 'Open Image...'], 'imageio:camera.png')

        gui.img.menu(['View', 'Refresh'])
        gui.img.menu(['View', 'Zoom In'])
        gui.img.menu(['View', 'Zoom Out'])
        gui.img.menu(['View', 'Zoom', 'Zoom 100%'])
        gui.img.menu(['View', 'Zoom', 'Zoom Fit'])
        gui.img.menu(['View', 'Zoom', 'Zoom Full'])
        gui.img.menu(['View', 'Zoom', 'Zoom Auto'])
        gui.img.menu(['View', 'Zoom', 'Zoom Exact...'], 50)

        gui.img.menu(['View', 'Default Offset & Gain'])
        gui.img.menu(['View', 'Set Current as Default'])
        gui.img.menu(['View', 'Offset & Gain...'], 0, 1, 2, 'grey')
        gui.img.menu(['View', 'Black & White...'], 10, 245, 'turbo')
        gui.img.menu(['View', 'Grey & Gain...'], 127, 4, 'jet')
        gui.img.menu(['View', 'Gain to Min-Max'])
        gui.img.menu(['View', 'Gain to Sigma', 'Gain to Sigma 1'])
        gui.img.menu(['View', 'Gain to Sigma', 'Gain to Sigma 2'])
        gui.img.menu(['View', 'Gain to Sigma', 'Gain to Sigma 3'])
        
        gui.img.menu(['View', 'HQ Zoom Out'])
        gui.img.menu(['View', 'Bind', 'Bind All Image Viewers'])
        gui.img.menu(['View', 'Bind', 'Unbind All Image Viewers'])
        gui.img.menu(['View', 'Bind', 'Absolute Zoom Link'])
        gui.img.menu(['View', 'Colormap...'], 'grey')
        gui.img.menu(['View', 'Colormap...'], 'clip')
        gui.img.menu(['View', 'Colormap...'], 'turbo')
        gui.img.menu(['View', 'Colormap...'], 'jet')
        gui.img.menu(['View', 'Colormap...'], 'invert')
        gui.img.menu(['View', 'Colormap...'], 'hot')
        gui.img.menu(['View', 'Colormap...'], 'cold')

        gui.img.menu(['View', 'Background Color...'], 58, 110, 165)
        gui.img.menu(['View', 'Selection Color...'], 255, 0, 0)
        gui.img.menu(['View', 'Value Format', 'Decimal'])
        gui.img.menu(['View', 'Value Format', 'Hex'])
        gui.img.menu(['View', 'Value Format', 'Binary'])
        gui.img.menu(['View', 'Show/Hide Profiles'])


    def test_menu_image_1(self):
        from gdesk import gui
        import imageio

        arr = imageio.imread('imageio:astronaut.png')

        gui.load_layout('image, levels & console')
        gui.img.select(1)
        imgpanid = gui.show(arr)
        gui.img.zoom_full()
        gui.menu_trigger('image', imgpanid, ['Image', 'Invert'])
        gui.menu_trigger('image', imgpanid, ['Image', 'Swap RGB | BGR'])
        gui.menu_trigger('image', imgpanid, ['Image', 'Adjust Lighting...'], 255, -1)
        gui.menu_trigger('image', imgpanid, ['Image', 'Swap RGB | BGR'])

        assert (arr == gui.vs).all()


    def test_menu_image_2(self):
        from gdesk import gui
        import imageio

        arr = imageio.imread('imageio:astronaut.png')

        gui.load_layout('image, levels & console')
        gui.img.select(1)
        gui.show(arr)

        gui.menu_trigger('image', None, ['Image', 'to Monochroom'])
        gui.menu_trigger('image', None, ['Edit', 'Show Prior Image'])
        gui.menu_trigger('image', None, ['Image', 'to Photometric Monochroom'])
        gui.menu_trigger('image', None, ['Edit', 'Show Prior Image'])

        assert (arr == gui.vs).all()


    def test_code_3(self):
        from gdesk import gui
        from pylab import plt

        plt.plot(gui.vs.mean(1))
        plt.grid(True)
        plt.title('Column Means')
        plt.show()
        plt.figure()
        plt.plot(gui.vs.mean(0))
        plt.grid(True)
        plt.title('Row Means')
        plt.show()

        answer = gui.question('Looks everything OK?')

        plt.close('all')
        gui.menu_trigger('image', GammaDeskSuite.panid, ['Edit', 'Show Prior Image'])


    def test_colors(self):
        import sys
        from gdesk import gui
        from pylab import plt

        gui.clc()
        plt.close('all')
        gui.load_layout('console')

        banner = r" _____                                      ______             _    " + "\n" \
                 r"|  __ \                                     |  _  \           | |   " + "\n" \
                 r"| |  \/  __ _  _ __ ___   _ __ ___    __ _  | | | |  ___  ___ | | __" + "\n" \
                 r"| | __  / _` || '_ ` _ \ | '_ ` _ \  / _` | | | | | / _ \/ __|| |/ /" + "\n" \
                 r"| |_\ \| (_| || | | | | || | | | | || (_| | | |/ / |  __/\__ \|   < " + "\n" \
                 r" \____/ \__,_||_| |_| |_||_| |_| |_| \__,_| |___/   \___||___/|_|\_" + "\\"

        fgs = [227, 227, 222, 217, 212, 207]
        for fg, line in zip(fgs, banner.splitlines()):
            bg = 236
            pline = '\033[1m\033[48;5;{0:03d}m\033[38;5;{1:03d}m'.format(bg, fg) + line + '\033[0m'
            print(pline)

        answer = gui.question('Does GammaDesk look nice in the console output?')
        assert answer


    def test_calc_pi_break(self):
        """Calculate pi digits and break after a few seconds"""
        import sys
        import time
        import threading
        from gdesk import shell, gui

        interpreter = shell.this_interpreter()
        ptid = threading.get_ident()
        panid = gui.console.selected()

        def delayedBreak():
            time.sleep(1)
            gui.redirects[threading.get_ident()] = ptid
            gui.menu_trigger('console', panid, ['Execution', 'Async Break'])

        threading.Thread(target=delayedBreak).start()

        def calcPi():
            q, r, t, k, n, l = 1, 0, 1, 1, 3, 3
            while True:
                if 4*q+r-t < n*t:
                    yield n
                    nr = 10*(r-n*t)
                    n  = ((10*(3*q+r))//t)-10*n
                    q  *= 10
                    r  = nr
                else:
                    nr = (2*q+r)*l
                    nn = (q*(7*k)+2+(r*l))//(t*l)
                    q  *= k
                    t  *= l
                    l  += 2
                    k += 1
                    n  = nn
                    r  = nr

        pi_digits = calcPi()
        i = 0

        print(f"{i:05d} ", end='')

        try:
            for d in pi_digits:
                sys.stdout.write(str(d))
                i += 1
                if (i % 64) == 0: print(f"\n{i:05d} ", end='')

        except KeyboardInterrupt:
            interpreter.break_sent = False


def suite():
    suite = unittest.TestSuite()
    suite.addTest(GammaDeskSuite('test_screenstate_1'))
    suite.addTest(GammaDeskSuite('test_small_loop_and_print'))
    suite.addTest(GammaDeskSuite('test_calc_pi_break'))
    suite.addTest(GammaDeskSuite('test_colors'))
    suite.addTest(GammaDeskSuite('test_menu_file'))
    suite.addTest(GammaDeskSuite('test_menu_canvas'))
    suite.addTest(GammaDeskSuite('test_menu_image_1'))
    suite.addTest(GammaDeskSuite('test_menu_image_2'))
    suite.addTest(GammaDeskSuite('test_code_3'))
    return suite




class DockBase(object):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        
    def get_container(self):
        candidate = self
        #TO DO, I can not import DockContainer here because it would cause circular imports
        #while not isinstance(candidate, DockContainer) ....
        #So this is the alternative
        while not (candidate.__class__.__name__ == 'DockContainer' or candidate is None):
            candidate = candidate.parent()
        return candidate

    def get_dock_box(self):
        candidate = self
        while not (candidate.__class__.__name__ in ['DockVBox', 'DockHBox'] or candidate is None):
            candidate = candidate.parent()
        return candidate                
        
    def detach(self, to_new_window=True):
        container = self.get_container()
        geo = self.geometry()
        return container.detach('layout', self.category, self.title, to_new_window, geo.width(), geo.height())
  